"""DataFrame on an Excel sheet."""

from __future__ import annotations

import re
from itertools import chain, takewhile
from typing import TYPE_CHECKING, overload

import numpy as np
import xlwings as xw
from pandas import DataFrame, MultiIndex, Series
from xlwings import Sheet

from xlviews import modify
from xlviews.axes import set_first_position
from xlviews.decorators import turn_off_screen_updating
from xlviews.element import Bar, Plot, Scatter
from xlviews.grid import FacetGrid
from xlviews.grouper import create_group_index
from xlviews.range import RangeCollection
from xlviews.style import set_alignment, set_frame_style, set_wide_column_style
from xlviews.table import Table
from xlviews.utils import iter_columns

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence
    from typing import Any, Literal

    from numpy.typing import ArrayLike, NDArray
    from xlwings import Range

    from xlviews.distframe import DistFrame
    from xlviews.grouper import Grouper
    from xlviews.statsframe import StatsFrame


class SheetFrame:
    """Data frame on an Excel sheet."""

    sheet: Sheet
    cell: Range
    name: str | None
    has_index: bool
    index_level: int
    columns_level: int
    columns_names: list[str] | None = None
    table: Table | None = None
    parent: SheetFrame | None
    children: list[SheetFrame]
    head: SheetFrame | None
    tail: SheetFrame | None = None
    stats: StatsFrame | None = None
    dist: DistFrame | None = None

    @turn_off_screen_updating
    def __init__(
        self,
        *args,
        name: str | None = None,
        sheet: Sheet | None = None,
        parent: SheetFrame | None = None,
        head: SheetFrame | None = None,
        data: DataFrame | None = None,
        index: bool = True,
        index_level: int = 1,
        columns_level: int = 1,
        style: bool = True,
        gray: bool = False,
        autofit: bool = True,
        number_format: str | None = None,
        font_size: int | None = None,
        table: bool = False,
        **kwargs,
    ) -> None:
        """Create a DataFrame on an Excel sheet.

        Args:
            sheet (Sheet): The sheet object.
            row, column (int): The position of the top-left cell.
            cell (Range): The Range of the top-left cell.
            data (DataFrame, optional): The DataFrame to write to the sheet.
            index (bool): Whether to output the index of the DataFrame.
            index_level (int): The depth of the index when importing data
                from the sheet.
            column_level (int): The depth of the columns when importing data
                from the sheet.
            parent (SheetFrame): The parent SheetFrame.
                The child SheetFrame is placed to the right of the parent SheetFrame.
                The parent SheetFrame represents the information extracted
                from the parent.
            head (SheetFrame): The upper SheetFrame.
                The child SheetFrame is placed below the upper SheetFrame.
                The upper SheetFrame represents the additional information
                of the parent.
            style (bool): Whether to decorate the SheetFrame.
            gray (bool): Whether to decorate the SheetFrame in gray.
            autofit (bool): Whether to autofit the SheetFrame.
            number_format (str): The number format of the SheetFrame.
            font_size (int): The font size of the SheetFrame.
        """
        self.name = name
        self.parent = parent
        self.children = []
        self.head = head

        if self.parent:  # Locate the child frame to the right of the parent frame.
            self.cell = self.parent.get_child_cell()
            self.parent.add_child_frame(self)

        elif self.head:  # Locate the child frame below the head frame.
            row_offset = len(self.head) + self.head.columns_level + 1
            self.cell = self.head.cell.offset(row_offset, 0)
            self.head.tail = self

        else:
            sheet = sheet or xw.sheets.active
            self.cell = sheet.range(*args)

        self.sheet = self.cell.sheet

        if data is not None:
            self.set_data(
                data,
                index=index,
                number_format=number_format,
                style=style,
                gray=gray,
                autofit=autofit,
                font_size=font_size,
                **kwargs,
            )
        else:
            self.set_data_from_sheet(
                index_level=index_level,
                columns_level=columns_level,
                number_format=number_format,
            )

        if table and not self.table:
            self.table = self.as_table()

    def set_data(
        self,
        data: DataFrame,
        *,
        index: bool = True,
        number_format: str | None = None,
        style: bool = True,
        gray: bool = False,
        autofit: bool = True,
        font_size: int | None = None,
        **kwargs,
    ) -> None:
        self.has_index = index
        self.index_level = len(data.index.names) if index else 0
        self.columns_level = len(data.columns.names)

        self.cell.options(DataFrame, index=index).value = data

        if index and data.columns.nlevels > 1 and data.index.nlevels == 1:
            self.columns_names = list(data.columns.names)
            self.cell.options(transpose=True).value = self.columns_names

        if number_format:
            self.set_number_format(number_format)

        if style:
            self.set_style(gray=gray, autofit=autofit, font_size=font_size, **kwargs)

        if self.head is None:
            self.set_adjacent_column_width(1)

        if self.name:
            book = self.sheet.book
            refers_to = "=" + self.cell.get_address(include_sheetname=True)
            book.names.add(self.name, refers_to)

    def set_data_from_sheet(
        self,
        *,
        index_level: int = 1,
        columns_level: int = 1,
        number_format: str | None = None,
    ) -> None:
        if self.name:
            book = self.sheet.book
            self.cell = book.names[self.name].refers_to_range
            self.sheet = self.cell.sheet

        self.has_index = bool(index_level)
        self.index_level = index_level
        self.columns_level = columns_level

        if self.columns_level > 1 and index_level == 1:
            start = self.cell
            end = start.offset(self.columns_level - 1)
            self.columns_names = self.sheet.range(start, end).value

        if number_format:
            self.set_number_format(number_format)

        for api in self.sheet.api.ListObjects:
            if api.Range.Row == self.row and api.Range.Column == self.column:
                self.table = Table(api=api, sheet=self.sheet)
                break

    def __len__(self) -> int:
        start = self.cell.offset(self.columns_level)
        cell = start

        while cell.value is not None:
            cell = cell.expand("down")[-1].offset(1)

        return cell.row - start.row

    @property
    def row(self) -> int:
        """Return the row of the top-left cell."""
        self.update_cell()
        return self.cell.row

    @property
    def column(self) -> int:
        """Return the column of the top-left cell."""
        self.update_cell()
        return self.cell.column

    @property
    def columns(self) -> list:
        """Return the column names."""
        if self.columns_level == 1:
            return self.expand("right").options(ndim=1).value or []

        if self.columns_names:
            idx = [tuple(self.columns_names)]
        elif self.has_index:
            start = self.cell.offset(self.columns_level - 1)
            end = start.offset(0, self.index_level - 1)
            idx = self.sheet.range(start, end).value or []
        else:
            idx = []

        cs = []
        for k in range(self.columns_level):
            rng = self.cell.offset(k, self.index_level).expand("right")
            cs.append(rng.options(ndim=1).value)
        cs = [tuple(c) for c in zip(*cs, strict=True)]

        return [*idx, *cs]

    @property
    def value_columns(self) -> list:
        return self.columns[self.index_level :]

    @property
    def index_columns(self) -> list[str | tuple[str, ...] | None]:
        return self.columns[: self.index_level]

    @property
    def wide_columns(self) -> list[str]:
        start = self.cell.offset(-1, self.index_level)
        end = start.offset(0, len(self.columns) - self.index_level - 1)
        cs = self.sheet.range(start, end).value or []
        return [c for c in cs if c]

    def __contains__(self, item: str | tuple) -> bool:
        return item in self.columns

    def __iter__(self) -> Iterator[str | tuple[str, ...] | None]:
        return iter(self.columns)

    @overload
    def index(
        self,
        column: str | tuple,
        *,
        relative: bool = False,
    ) -> int | tuple[int, int]: ...

    @overload
    def index(
        self,
        column: list[str | tuple],
        *,
        relative: bool = False,
    ) -> list[int | tuple[int, int]]: ...

    @overload
    def index(
        self,
        column: dict,
        *,
        relative: bool = False,
    ) -> tuple[int, int]: ...

    def index(
        self,
        column: str | tuple | list[str | tuple] | dict,
        *,
        relative: bool = False,
    ) -> int | tuple[int, int] | list[int | tuple[int, int]]:
        """Return the column index (1-indexed).

        If the column is a hierarchical index and the column name is specified,
        return the row index. If relative is True, return the relative position
        from `self.cell`.
        """
        if isinstance(column, list):
            return [self.index(c, relative=relative) for c in column]

        if isinstance(column, dict):
            return self.index_dict(column, relative=relative)

        if self.columns_names and isinstance(column, str):
            return self.index_row(column, relative=relative)

        columns = self.columns

        if column in columns:
            offset = 1 if relative else self.column
            return columns.index(column) + offset

        return self.index_wide(column, relative=relative)

    def index_row(self, column: str, *, relative: bool = False) -> int:
        if not self.columns_names:
            raise NotImplementedError

        row = self.columns_names.index(column)
        return row + 1 if relative else row + self.row

    @overload
    def index_wide(self, column: str, *, relative: bool = False) -> tuple[int, int]: ...

    @overload
    def index_wide(
        self,
        column: tuple[str, str | float],
        *,
        relative: bool = False,
    ) -> int: ...

    def index_wide(
        self,
        column: str | tuple[str, str | float],
        *,
        relative: bool = False,
    ) -> tuple[int, int] | int:
        value_columns = self.value_columns

        cell_start = self.cell.offset(-1, self.index_level)
        cell_end = cell_start.offset(0, len(value_columns) - 1)
        names = self.sheet.range(cell_start, cell_end).options(ndim=1).value or []

        name = column[0] if isinstance(column, tuple) else column
        start = names.index(name)
        end = len(list(takewhile(lambda n: n is None, names[start + 1 :]))) + start

        offset = self.index_level + (1 if relative else self.cell.column)

        if isinstance(column, str):
            return start + offset, end + offset

        values = value_columns[start : end + 1]
        return values.index(column[1]) + start + offset

    def index_dict(self, column: dict, *, relative: bool = False) -> tuple[int, int]:
        if not self.columns_names:
            raise NotImplementedError

        by = list(column.keys())
        key = tuple(column.values())

        index = self.groupby(by)[key]

        if len(index) != 1 or len(index[0]) != 2:
            raise NotImplementedError

        offset = -self.column + 1 if relative else 0
        return index[0][0] + offset, index[0][1] + offset

    def update_cell(self) -> None:  # for bug fix in cell.row, cell.column, cell.expand
        self.cell = self.cell.offset(0, 0)

    def expand(self, mode: str = "table") -> Range:
        self.update_cell()
        return self.cell.expand(mode)

    @property
    def data(self) -> DataFrame:
        """Return the data as a DataFrame."""
        if self.cell.value is None and self.columns_level > 1:
            rng = self.cell.offset(self.columns_level - 1).expand()
            rng = rng.options(DataFrame, index=self.index_level, header=1)
            df = rng.value
            df.columns = MultiIndex.from_tuples(self.value_columns)
            return df

        rng = self.expand()
        rng = rng.options(DataFrame, index=self.index_level, header=self.columns_level)
        df = rng.value

        if not isinstance(df, DataFrame):
            raise NotImplementedError

        if self.columns_names:
            df.index.name = None
            df.columns.names = self.columns_names

        return df

    @property
    def visible_data(self) -> DataFrame:
        self.update_cell()
        start = self.cell.offset(1, 0)
        end = start.offset(len(self) - 1, len(self.columns) - 1)
        range_ = self.sheet.range(start, end)
        data = range_.api.SpecialCells(xw.constants.CellType.xlCellTypeVisible)
        value = [row.Value[0] for row in data.Rows]
        df = DataFrame(value, columns=self.columns)

        if self.has_index and self.index_level:
            df = df.set_index(list(df.columns[: self.index_level]))

        return df

    def as_table(
        self,
        *,
        const_header: bool = True,
        autofit: bool = True,
        style: bool = True,
    ) -> Table:
        if self.columns_level != 1:
            raise NotImplementedError

        self.set_alignment("left")

        end = self.cell.offset(len(self), len(self.columns) - 1)
        rng = self.sheet.range(self.cell, end)

        table = Table(rng, autofit=autofit, const_header=const_header, style=style)
        self.table = table

        return table

    def unlist(self) -> None:
        if self.table:
            self.table.unlist()
            self.table = None

    @overload
    def range(
        self,
        column: str | tuple | dict | None = None,
        start: int | Literal[False] | None = None,
        end: int | None = None,
    ) -> Range: ...

    @overload
    def range(
        self,
        column: str | tuple | dict,
        start: list[tuple[int, int]],
        end: int | None = None,
    ) -> RangeCollection: ...

    def range(
        self,
        column: str | tuple | dict | None = None,
        start: int | Literal[False] | list[tuple[int, int]] | None = None,
        end: int | None = None,
    ) -> Range | RangeCollection:
        """Return the range of the column.

        If the column is a hierarchical index and the column name is specified,
        return the range of the column.

        Args:
            column (str or tuple or dict, optional): The name of the column.
                If omitted, return the range of the entire SheetFrame.
                If a dict is specified, filter by the hierarchical column.
            start (int, optional):
                - None: entire row data without column row
                - 0: first row
                - -1: column row
                - False: entire row with column row
                - other: specified row
            end (int, optional):
                - None : same as start.
                - other: specified row
        """
        if column is None:
            return self.range_all()

        if isinstance(start, list):
            return RangeCollection(self.range(column, s, e) for s, e in start)

        if column == "index":
            return self.range_index(start, end)

        return self.range_column(column, start, end)

    def range_all(self) -> Range:
        start = self.cell
        row_offset = self.columns_level + len(self) - 1
        column_offset = self.index_level + len(self.value_columns) - 1
        end = start.offset(row_offset, column_offset)
        return self.sheet.range(start, end)

    def range_index(
        self,
        start: int | Literal[False] | None = None,
        end: int | None = None,
    ) -> Range:
        """Return the range of the index."""
        if not self.index_level:
            raise NotImplementedError

        if self.columns_names and (start in [-1, None] or start is False):
            raise ValueError("index start must be a specific row")

        c = self.index_level - 1
        match start:
            case False:  # entire row with column row
                cell_start = self.cell
                cell_end = cell_start.offset(self.columns_level + len(self) - 1, c)

            case 0:  # first row
                cell_start = self.cell.offset(self.columns_level)
                cell_end = cell_start.offset(0, c)

            case -1:  # column row
                cell_start = self.cell
                cell_end = cell_start.offset(self.columns_level - 1, c)

            case None:  # entire row data without column row
                cell_start = self.cell.offset(self.columns_level)
                cell_end = cell_start.offset(len(self) - 1, c)

            case _:
                column = self.cell.column
                cell_start = self.sheet.range(start, column)
                cell_end = self.sheet.range(end or start, column + c)

        return self.sheet.range(cell_start, cell_end)

    def range_column(  # noqa: C901
        self,
        column: str | tuple | dict,
        start: int | Literal[False] | None = None,
        end: int | None = None,
    ) -> Range:
        if self.columns_names and isinstance(column, str):
            raise NotImplementedError

        if start is False:  # entire row with column row
            header = self.range_column(column, -1)
            values = self.range_column(column)
            return self.sheet.range(header[0], values[-1])

        index = self.index(column)

        match start:
            case 0:  # first row
                start = self.row + self.columns_level

            case -1:  # column row
                start = end = self.row
                if isinstance(column, tuple) and self.columns_level == 1:
                    start -= 1  # wide column
                else:
                    end += self.columns_level - 1
                    if isinstance(index, tuple):
                        start -= 1

            case None:  # entire row data without column row
                start = self.row + self.columns_level
                end = start + len(self) - 1

        if isinstance(index, tuple):  # wide column
            column_start, column_end = index
        else:
            column_start = column_end = index

        cell_start = self.sheet.range(start, column_start)

        if end is None:
            if not isinstance(index, tuple):
                return cell_start

            # wide column
            cell_end = cell_start.offset(0, column_end - column_start)
            return self.sheet.range(cell_start, cell_end)

        cell_end = self.sheet.range(end, column_end)
        return self.sheet.range(cell_start, cell_end)

    def first_range(self, column: str | tuple | dict) -> Range:
        return self.range(column, 0)

    def __repr__(self) -> str:
        return repr(self.range()).replace("<Range ", "<SheetFrame ")

    def __str__(self) -> str:
        return str(self.range()).replace("<Range ", "<SheetFrame ")

    def rename(self, columns: dict[str, str]) -> None:
        """Rename the columns of the SheetFrame."""
        for old, new in columns.items():
            self.range(old, -1).value = new

    def drop_duplicates(self, column: str | tuple | Iterable[str | tuple]) -> None:
        columns = [column] if isinstance(column, str | tuple) else list(column)

        for column in columns:
            for cell in reversed(self.range(column)[1:]):
                if cell.value == cell.offset(-1).value:
                    cell.value = None

    def get_address(
        self,
        column: str | tuple | dict,
        *,
        formula: bool = False,
        **kwargs,
    ) -> list[str]:
        """Return the address list of the column.

        Args:
            column (str or tuple or dict): The name of the column.
            formula (bool, optional): Whether to add '=' to the beginning
                of the address.
            kwargs Keyword arguments for the `Range.get_address` method.

        Returns:
            list[str]: The address list of the column.
        """
        addresses = []
        for cell in self.range(column):
            addresses.append(cell.get_address(**kwargs))

        if formula:
            addresses = ["=" + address for address in addresses]

        return addresses

    def add_column(self, column: str, value: Any | None = None) -> Range:
        column_int = self.column + len(self.columns)
        cell = self.sheet.range(self.row, column_int)
        cell.value = column

        rng = self.range(column)

        if value is not None:
            rng.options(transpose=True).value = value

        return rng

    def add_formula_column(
        self,
        rng: Range | str,
        formula: str,
        *,
        number_format: str | None = None,
        autofit: bool = False,
    ) -> None:
        """Add a formula column.

        Args:
            rng (Range): The range of the column.
            formula (str or tuple): The formula.
            number_format (str, optional): The number format.
            autofit (bool): Whether to autofit the width.
        """
        columns = self.columns
        wide_columns = self.wide_columns

        if isinstance(rng, str):
            if rng not in columns + wide_columns:
                rng = self.add_column(rng)
            else:
                rng = self.range(rng)

        refs = {}
        for m in re.finditer(r"{(.+?)}", formula):
            column = m.group(1)

            if column in columns:
                ref = self.range(column, 0)
                ref = ref.get_address(row_absolute=False)

            elif column in wide_columns:
                ref = self.range(column, -1)[0].offset(1)
                ref = ref.get_address(column_absolute=False)

            else:
                ref = self.range(column, 0)[0]
                ref = ref.get_address(column_absolute=False, row_absolute=False)

            refs[column] = ref

        rng.value = formula.format(**refs)

        if number_format:
            rng.number_format = number_format

        if autofit:
            rng = rng.sheet.range(rng[0].offset(-1), rng[-1])
            rng.autofit()

    def add_wide_column(
        self,
        column: str,
        values: Iterable[str | float],
        *,
        number_format: str | None = None,
        autofit: bool = True,
        style: bool = False,
    ) -> Range:
        """Add a wide column.

        Args:
            column (str): The name of the wide column.
            values (iterable): The values to be expanded horizontally.
            number_format (str, optional): The number format.
            autofit (bool): Whether to autofit the width.
            style (bool): Whether to style the column.
        """
        if self.columns_level != 1:
            raise NotImplementedError

        rng = self.cell.offset(0, len(self.columns))
        values_list = list(values)
        rng.value = values_list

        header = rng.offset(-1)
        header.value = column

        set_alignment(header, horizontal_alignment="left")

        rng = rng.sheet.range(rng, rng.offset(0, len(values_list)))
        if number_format:
            rng.number_format = number_format

        if autofit:
            self.range(column, -1).autofit()

        if style:
            self.set_style()

        return rng[0].offset(1)

    @overload
    def __getitem__(self, column: str | tuple) -> Series: ...

    @overload
    def __getitem__(self, column: Sequence[str | tuple]) -> DataFrame: ...

    def __getitem__(
        self,
        column: str | tuple | Sequence[str | tuple],
    ) -> Series | DataFrame:
        """Return the column data.

        If column is a string, return a Series. If column is a list,
        return a DataFrame. The index is ignored.
        """
        if isinstance(column, list):
            return DataFrame({c: self[c] for c in column})

        if isinstance(column, str | tuple):
            row = self.row + self.columns_level
            name, column_index = column, self.index(column)
            start = self.sheet.range(row, column_index)

            if len(self) == 1:
                array = [start.value]
            else:
                end = start.offset(len(self) - 1, 0)
                rng = self.sheet.range(start, end)
                array = rng.options(np.array).value

            return Series(array, name=name)

        raise NotImplementedError

    def __setitem__(self, column: str | tuple, value: ArrayLike) -> None:
        if column in self:
            rng = self.range(column)
        elif isinstance(column, str):
            rng = self.add_column(column)
        else:
            raise NotImplementedError

        if isinstance(value, str) and value.startswith("="):
            self.add_formula_column(rng, value)
        else:
            rng.options(transpose=True).value = value

    def select(self, **kwargs) -> NDArray[np.bool_]:
        """Return the selection of the SheetFrame.

        Keyword arguments are column names and values. The conditions are as follows:
           - list : the specified elements are selected.
           - tuple : the range of the value.
           - other : the value is selected if it matches.
        """

        def filter_(
            sel: NDArray[np.bool_],
            array: Series,
            value: str | list | tuple,
        ) -> None:
            if isinstance(value, list):
                sel &= array.isin(value)
            elif isinstance(value, tuple):
                sel &= (array >= value[0]) & (array <= value[1])
            else:
                sel &= array == value

        if self.columns_names is None:
            # vertical selection
            sel = np.ones(len(self), dtype=bool)

            for key, value in kwargs.items():
                filter_(sel, self[key], value)

            return sel

        # horizontal selection
        columns = self.value_columns
        sel = np.ones(len(columns), dtype=bool)
        df = DataFrame(columns, columns=self.columns_names)

        for key, value in kwargs.items():
            filter_(sel, df[key], value)

        return sel

    def ranges(
        self,
        sel: Sequence[bool] | NDArray[np.bool_] | None = None,
        **kwargs,
    ) -> Iterator[Range]:
        if sel is None:
            n = len(self) if self.columns_names is None else len(self.value_columns)
            sel = np.ones(n, dtype=bool)
        elif not isinstance(sel, np.ndarray):
            sel = np.array(sel, dtype=bool)

        if kwargs:
            sel &= self.select(**kwargs)

        if self.columns_names is None:
            yield from self.ranges_row(sel)
        else:
            yield from self.ranges_column(sel)

    def ranges_row(self, sel: NDArray[np.bool_]) -> Iterator[Range]:
        offset = self.row + self.columns_level
        start = self.column + self.index_level
        end = start + len(self.value_columns) - 1

        for index in range(len(self)):
            if not sel[index]:
                continue

            yield self.sheet.range((index + offset, start), (index + offset, end))

    def ranges_column(self, sel: NDArray[np.bool_]) -> Iterator[Range]:
        offset = self.column + self.index_level
        start = self.row + self.columns_level
        end = start + len(self) - 1

        for index in range(len(self.value_columns)):
            if not sel[index]:
                continue

            yield self.sheet.range((start, index + offset), (end, index + offset))

    def groupby(
        self,
        by: str | list[str] | None,
    ) -> dict[tuple, list[tuple[int, int]]]:
        """Group by the specified column and return the group key and row number."""
        if not by:
            if self.columns_names is None:
                start = self.row + self.columns_level
                end = start + len(self) - 1
                return {(): [(start, end)]}

            start = self.column + 1
            end = start + len(self.value_columns) - 1
            return {(): [(start, end)]}

        if self.columns_names is None:
            if isinstance(by, list) or ":" in by:
                by = list(iter_columns(self, by))
            values = self[by]

        else:
            df = DataFrame(self.value_columns, columns=self.columns_names)
            values = df[by]

        index = create_group_index(values)

        if self.columns_names is None:
            offset = self.row + self.columns_level  # vertical
        else:
            offset = self.column + self.index_level  # horizontal

        return {k: [(x + offset, y + offset) for x, y in v] for k, v in index.items()}

    def grouper(self, by: str | list[str] | None) -> Grouper:
        from xlviews.grouper import Grouper

        return Grouper(self, by)

    def get_number_format(self, column: str | tuple) -> str:
        return self.range(column, 0).number_format

    def set_number_format(
        self,
        number_format: str | dict | None = None,
        *,
        autofit: bool = False,
        **columns_format,
    ) -> None:
        if isinstance(number_format, str):
            start = self.cell.offset(self.columns_level, self.index_level)
            rng = self.sheet.range(start, self.range()[-1])
            rng.number_format = number_format
            if autofit:
                rng.autofit()
            return

        if isinstance(number_format, dict):
            columns_format.update(number_format)

        for column in chain(self.columns, self.wide_columns):
            if not column:
                continue

            for pattern, number_format in columns_format.items():
                column_name = column if isinstance(column, str) else column[0]

                if re.match(pattern, column_name):
                    rng = self.range(column)
                    rng.number_format = number_format
                    if autofit:
                        rng.autofit()
                    break

    def set_style(self, *, gray: bool = False, **kwargs) -> None:
        set_frame_style(self, gray=gray, **kwargs)
        set_wide_column_style(self, gray=gray)

    def set_alignment(self, alignment: str) -> None:
        start = self.cell
        end = start.offset(0, len(self.columns) - 1)
        rng = self.sheet.range(start, end)
        set_alignment(rng, alignment)

    def set_adjacent_column_width(self, width: float) -> None:
        """Set the width of the adjacent empty column."""
        column = self.column + len(self.columns)
        self.sheet.range(1, column).column_width = width

    def add_child_frame(self, child: SheetFrame) -> None:
        """Add a child SheetFrame."""
        self.children.append(child)
        child.parent = self

    def get_child_cell(self) -> Range:
        """Get the cell of the child SheetFrame."""
        offset = len(self.columns) + 1
        offset += sum(len(child.columns) + 1 for child in self.children)
        return self.cell.offset(0, offset)

    def get_adjacent_cell(self, offset: int = 0) -> Range:
        """Get the adjacent cell of the SheetFrame."""
        if self.children:
            return self.get_child_cell()

        return self.cell.offset(0, len(self.columns) + offset + 1)

    def move(self, count: int, direction: str = "down", width: int = 0) -> Range:
        return modify.move(self, count, direction, width)

    def delete(self, direction: str = "up", *, entire: bool = False) -> None:
        return modify.delete(self, direction, entire=entire)

    @turn_off_screen_updating
    def copy(self, *args, **kwargs) -> SheetFrame:
        return modify.copy(self, *args, **kwargs)

    def distframe(self, *args, **kwargs) -> DistFrame:
        from xlviews.distframe import DistFrame

        self.dist = DistFrame(self, *args, **kwargs)
        return self.dist

    def statsframe(self, *args, **kwargs) -> StatsFrame:
        from xlviews.statsframe import StatsFrame

        self.stats = StatsFrame(self, *args, **kwargs)
        return self.stats

    def set_chart_position(self, pos: str = "right") -> None:
        set_first_position(self, pos=pos)

    def scatter(self, *args, **kwargs) -> Scatter:
        return Scatter(*args, data=self, **kwargs)

    def plot(self, *args, **kwargs) -> Plot:
        return Plot(*args, data=self, **kwargs)

    def bar(self, *args, **kwargs) -> Bar:
        return Bar(*args, data=self, **kwargs)

    def grid(self, *args, **kwargs) -> FacetGrid:
        return FacetGrid(self, *args, **kwargs)



import pytest

from xlviews.grouper import Grouper
from xlviews.range import RangeCollection
from xlviews.sheetframe import SheetFrame


@pytest.mark.parametrize(
    ("by", "n"),
    [(None, 1), ("x", 2), (["x", "y"], 4), (["x", "y", "z"], 20)],
)
def test_by(sf: SheetFrame, by, n):
    gr = Grouper(sf, by)
    assert len(gr.grouped) == n


@pytest.fixture(scope="module")
def gr(sf: SheetFrame):
    return Grouper(sf, ["x", "y"])


def test_group_key(gr: Grouper):
    keys = list(gr.grouped.keys())
    assert keys == [("a", "c"), ("a", "d"), ("b", "c"), ("b", "d")]


def test_ranges_len(gr: Grouper):
    assert len(list(gr.ranges("a"))) == 4


@pytest.mark.parametrize(("column", "c"), [("x", "C"), ("y", "D")])
def test_first_ranges(gr: Grouper, column, c):
    rs = [r.get_address() for r in gr.first_ranges(column)]
    assert rs == [f"${c}$4", f"${c}$8", f"${c}$12", f"${c}$16"]


@pytest.mark.parametrize(("column", "c"), [("a", "F"), ("b", "G"), ("c", "H")])
@pytest.mark.parametrize(
    ("k", "a"),
    [
        (0, "${c}$4:${c}$7,${c}$20:${c}$23"),
        (1, "${c}$8:${c}$11"),
        (2, "${c}$12:${c}$15"),
        (3, "${c}$16:${c}$19"),
    ],
)
def test_ranges(gr: Grouper, column, c, k: int, a):
    rc = list(gr.ranges(column))[k]
    assert isinstance(rc, RangeCollection)
    assert rc.get_address() == a.format(c=c)



import numpy as np
import pytest
from pandas import DataFrame
from xlwings import Range, Sheet

from xlviews.formula import NONCONST_VALUE
from xlviews.range import RangeCollection


@pytest.fixture(scope="module")
def df():
    return DataFrame({"a": [1, 1, 1, 1], "b": [2, 2, 3, 3], "c": [4, 4, 4, 4]})


@pytest.fixture(scope="module")
def rng(df: DataFrame, sheet_module: Sheet):
    rng = sheet_module.range("B3")
    rng.options(DataFrame, header=1, index=False).value = df
    return rng.expand()


def test_range_address(rng: Range):
    assert rng.get_address() == "$B$3:$D$7"


@pytest.mark.parametrize("k", range(3))
def test_range_value(rng: Range, df: DataFrame, k: int):
    value = rng.options(transpose=True).value
    assert isinstance(value, list)
    assert len(value) == 3
    np.testing.assert_array_equal(df[value[k][0]], value[k][1:])


@pytest.fixture(scope="module")
def column(rng: Range):
    start = rng[0].offset(1)
    end = start.offset(3)
    return rng.sheet.range(start, end)


def test_column(column: Range):
    assert column.get_address() == "$B$4:$B$7"


@pytest.fixture(scope="module")
def const_header(rng: Range):
    end = rng[0].expand("right")
    return rng.sheet.range(rng[0], end).offset(-1)


def test_header(const_header: Range):
    assert const_header.get_address() == "$B$2:$D$2"


@pytest.mark.parametrize(("k", "value"), [(0, 1), (1, NONCONST_VALUE), (2, 4)])
def test_const(column: Range, const_header: Range, k, value):
    from xlviews.formula import const

    const_header.value = const(column, "=")
    assert const_header.value[k] == value


@pytest.fixture(scope="module")
def ranges(sheet_module: Sheet):
    rng = sheet_module.range("B100")
    rng.options(transpose=True).value = [1, 2, 3, 4, 0, 5, 6, 7, 8, 9, 10]
    rng1 = rng.expand("down")
    sheet_module.range("B104").value = None

    rng = sheet_module.range("C100")
    rng.options(transpose=True).value = [11, 12, 13, 14, 15, 16, 17, 0, 18, 19, 20]
    rng2 = rng.expand("down")
    sheet_module.range("C107").value = None

    return [rng1, rng2]


def test_ranges(ranges: list[Range]):
    assert ranges[0].get_address() == "$B$100:$B$110"
    assert ranges[1].get_address() == "$C$100:$C$110"


@pytest.mark.parametrize("apply", [list, RangeCollection])
def test_aggregate_value(ranges: list[Range], apply):
    from xlviews.formula import aggregate

    x = aggregate("count", apply(ranges))
    assert x == "AGGREGATE(2,7,$B$100:$B$110,$C$100:$C$110)"


FUNC_VALUES = [
    ("count", 20),
    ("sum", 210),
    ("min", 1),
    ("max", 20),
    ("mean", 10.5),
    ("median", 10.5),
    ("std", np.std(range(1, 21))),
    ("soa", np.std(range(1, 21)) / 10.5),
]


@pytest.mark.parametrize(("func", "value"), FUNC_VALUES)
@pytest.mark.parametrize("apply", [list, RangeCollection])
def test_aggregate_str(ranges: list[Range], apply, func, value):
    from xlviews.formula import aggregate

    formula = aggregate(func, apply(ranges))
    cell = ranges[0].sheet.range("D100")
    cell.value = "=" + formula
    assert cell.value == value


@pytest.mark.parametrize(("func", "value"), FUNC_VALUES)
@pytest.mark.parametrize("apply", [list, RangeCollection])
def test_aggregate_range(ranges: list[Range], apply, func, value):
    from xlviews.formula import aggregate

    ref = ranges[0].sheet.range("E100")
    ref.value = func
    formula = aggregate(ref, apply(ranges))
    cell = ranges[0].sheet.range("D100")
    cell.value = "=" + formula
    assert cell.value == value



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame


def test_wrap_wrap():
    from xlviews.statsframe import get_wrap

    assert get_wrap(wrap="wrap") == "wrap"
    assert get_wrap(wrap={"a": "wrap"}) == {"a": "wrap"}


@pytest.mark.parametrize(
    ("kwarg", "value"),
    [("na", "IFERROR({},NA())"), ("null", 'IFERROR({},"")')],
)
def test_wrap_true(kwarg, value):
    from xlviews.statsframe import get_wrap

    assert get_wrap(**{kwarg: True}) == value  # type: ignore


def test_wrap_list():
    from xlviews.statsframe import get_wrap

    x = get_wrap(na=["a", "b"], null="c")
    assert isinstance(x, dict)
    assert x["a"] == "IFERROR({},NA())"
    assert x["b"] == "IFERROR({},NA())"
    assert x["c"] == 'IFERROR({},"")'


def test_wrap_none():
    from xlviews.statsframe import get_wrap

    assert get_wrap() is None


def test_func_none():
    from xlviews.statsframe import get_func

    func = ["count", "max", "mean", "median", "min", "soa"]
    assert sorted(get_func(None)) == func


def test_func_str():
    from xlviews.statsframe import get_func

    assert get_func("count") == ["count"]


@pytest.mark.parametrize("func", [["count"], {"a": "count"}])
def test_func_else(func):
    from xlviews.statsframe import get_func

    assert get_func(func) == func


@pytest.mark.parametrize(
    ("funcs", "n"),
    [(["mean"], 4), (["min", "max", "median"], 12), ({"a": "count"}, 4)],
)
def test_length(sf_parent: SheetFrame, funcs, n):
    from xlviews.statsframe import get_length

    assert get_length(sf_parent, ["x", "y"], funcs) == n


def test_length_none_list(sf_parent: SheetFrame):
    from xlviews.statsframe import get_length

    assert get_length(sf_parent, [], ["min", "max"]) == 2


def test_length_none_dict(sf_parent: SheetFrame):
    from xlviews.statsframe import get_length

    assert get_length(sf_parent, [], {"a": "mean"}) == 1


def test_has_header(sf_parent: SheetFrame):
    from xlviews.statsframe import has_header

    assert has_header(sf_parent)


def test_move_down(sheet: Sheet):
    from xlviews.statsframe import move_down

    df = DataFrame([[1, 2, 3], [4, 5, 6]], columns=["a", "b", "c"])
    sf = SheetFrame(3, 3, data=df, style=False, sheet=sheet)
    assert sheet["D3:F3"].value == ["a", "b", "c"]
    assert sheet["D2:F2"].value == [None, None, None]
    assert move_down(sf, 3) == 3
    assert sheet["D6:F6"].value == ["a", "b", "c"]
    assert sheet["D5:F5"].value == [None, None, None]
    assert sf.row == 6


def test_move_down_header(sheet: Sheet):
    from xlviews.statsframe import move_down

    df = DataFrame([[1, 2, 3], [4, 5, 6]], columns=["a", "b", "c"])
    sf = SheetFrame(3, 3, data=df, style=False, sheet=sheet)
    sheet["D2"].value = "x"
    assert sheet["D3:F3"].value == ["a", "b", "c"]
    assert sheet["D2:F2"].value == ["x", None, None]
    assert move_down(sf, 3) == 4
    assert sheet["D7:F7"].value == ["a", "b", "c"]
    assert sheet["D6:F6"].value == ["x", None, None]
    assert sf.row == 7



import numpy as np
import pytest
from pandas import DataFrame, MultiIndex, Series
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame


@pytest.fixture(scope="module")
def df():
    df = DataFrame(
        {
            "x": [1, 1, 1, 1, 2, 2, 2, 2],
            "y": [1, 1, 2, 2, 1, 1, 2, 2],
            "z": [1, 1, 1, 2, 2, 1, 1, 1],
            "c": [1, 2, 3, 4, 5, 6, 7, 8],
            "d": [11, 12, 13, 14, 15, 16, 17, 18],
            "e": [21, 22, 23, 24, 25, 26, 27, 28],
            "f": [31, 32, 33, 34, 35, 36, 37, 38],
        },
    )
    df = df.set_index(["x", "y", "z"])
    x = [("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
    df.columns = MultiIndex.from_tuples(x, names=["a", "b"])
    return df


def test_df(df: DataFrame):
    assert len(df) == 8
    assert df.shape == (8, 4)

    x = [("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
    assert df.columns.to_list() == x
    assert df.columns.names == ["a", "b"]
    assert isinstance(df.columns, MultiIndex)

    x = [(1, 1, 1), (1, 1, 1), (1, 2, 1), (1, 2, 2), (2, 1, 2), (2, 1, 1)]
    x += [(2, 2, 1), (2, 2, 1)]
    assert df.index.to_list() == x
    assert df.index.names == ["x", "y", "z"]
    assert isinstance(df.index, MultiIndex)


@pytest.fixture(scope="module")
def sf(df: DataFrame, sheet_module: Sheet):
    return SheetFrame(5, 3, data=df, style=False, sheet=sheet_module)


def test_value_column(sf: SheetFrame):
    assert sf.cell.value is None
    v = sf.cell.offset(0, 3).expand().options(ndim=2).value
    assert len(v) == 10
    assert v[0] == ["a1", "a1", "a2", "a2"]
    assert v[1] == ["b1", "b2", "b1", "b2"]
    assert v[2] == [1, 11, 21, 31]
    assert v[-1] == [8, 18, 28, 38]


def test_value_values(sf: SheetFrame):
    v = sf.cell.offset(1, 0).expand().options(ndim=2).value
    assert len(v) == 9
    assert v[0] == ["x", "y", "z", "b1", "b2", "b1", "b2"]
    assert v[1] == [1, 1, 1, 1, 11, 21, 31]
    assert v[2] == [1, 1, 1, 2, 12, 22, 32]
    assert v[-1] == [2, 2, 1, 8, 18, 28, 38]


def test_init(sf: SheetFrame, sheet_module: Sheet):
    assert sf.cell.get_address() == "$C$5"
    assert sf.row == 5
    assert sf.column == 3
    assert sf.sheet.name == sheet_module.name
    assert sf.index_level == 3
    assert sf.columns_level == 2
    assert sf.columns_names is None


def test_set_data_from_sheet(sf: SheetFrame):
    sf.set_data_from_sheet(index_level=2, columns_level=2)
    assert sf.has_index is True
    assert sf.index_columns == ["x", "y"]
    c = [(None, "z"), ("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
    assert sf.value_columns == c
    sf.set_data_from_sheet(index_level=3, columns_level=2)
    assert sf.has_index is True
    assert sf.index_columns == ["x", "y", "z"]
    c = [("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
    assert sf.value_columns == c


def test_len(sf: SheetFrame):
    assert len(sf) == 8


def test_columns(sf: SheetFrame):
    i = "x", "y", "z"
    c = ("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")
    assert sf.columns == [*i, *c]


def test_value_columns(sf: SheetFrame):
    c = [("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
    assert sf.value_columns == c


def test_index_columns(sf: SheetFrame):
    assert sf.index_columns == ["x", "y", "z"]


def test_init_index_false(df: DataFrame, sheet: Sheet):
    sf = SheetFrame(2, 3, data=df, index=False, style=False, sheet=sheet)

    assert sf.index_level == 0
    c = [("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
    assert sf.columns == c


def test_contains(sf: SheetFrame):
    assert "x" in sf
    assert "z" in sf
    assert ("a1", "b1") in sf
    assert "a1" not in sf


def test_iter(sf: SheetFrame):
    i = "x", "y", "z"
    c = ("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")
    assert list(sf) == [*i, *c]


@pytest.mark.parametrize(
    ("column", "relative", "index"),
    [
        ("x", True, 1),
        ("z", False, 5),
        (("a1", "b1"), True, 4),
        (("a1", "b1"), False, 6),
        (["y", ("a2", "b1")], True, [2, 6]),
        (["x", ("a2", "b2")], False, [3, 9]),
    ],
)
def test_index(sf: SheetFrame, column, relative, index):
    assert sf.index(column, relative=relative) == index


def test_index_error(sf: SheetFrame):
    with pytest.raises(ValueError, match="'a' is not in list"):
        sf.index("a")


def test_data(sf: SheetFrame, df: DataFrame):
    df_ = sf.data
    np.testing.assert_array_equal(df_.index, df.index)
    np.testing.assert_array_equal(df_.index.names, df.index.names)
    np.testing.assert_array_equal(df_.columns, df.columns)
    np.testing.assert_array_equal(df_, df)
    assert df_.index.name == df.index.name


def test_range_all(sf: SheetFrame):
    assert sf.range_all().get_address() == "$C$5:$I$14"
    assert sf.range().get_address() == "$C$5:$I$14"


@pytest.mark.parametrize(
    ("start", "end", "address"),
    [
        (False, None, "$C$5:$E$14"),
        (-1, None, "$C$5:$E$6"),
        (0, None, "$C$7:$E$7"),
        (None, None, "$C$7:$E$14"),
        (20, None, "$C$20:$E$20"),
        (20, 100, "$C$20:$E$100"),
    ],
)
def test_range_index(sf: SheetFrame, start, end, address):
    assert sf.range_index(start, end).get_address() == address
    assert sf.range("index", start, end).get_address() == address


@pytest.mark.parametrize(
    ("column", "start", "end", "address"),
    [
        ("x", -1, None, "$C$5:$C$6"),
        ("y", 0, None, "$D$7"),
        ("z", None, None, "$E$7:$E$14"),
        (("a2", "b2"), False, None, "$I$5:$I$14"),
        (("a1", "b1"), -1, None, "$F$5:$F$6"),
        (("a1", "b2"), 0, None, "$G$7"),
        (("a2", "b1"), None, None, "$H$7:$H$14"),
        ("y", 30, None, "$D$30"),
        (("a1", "b2"), 30, 40, "$G$30:$G$40"),
    ],
)
def test_range_column(sf: SheetFrame, column, start, end, address):
    assert sf.range_column(column, start, end).get_address() == address
    assert sf.range(column, start, end).get_address() == address


def test_getitem_str(sf: SheetFrame):
    s = sf["x"]
    assert isinstance(s, Series)
    assert s.name == "x"
    np.testing.assert_array_equal(s, [1, 1, 1, 1, 2, 2, 2, 2])


def test_getitem_list(sf: SheetFrame):
    df = sf[["z", ("a1", "b1")]]
    assert isinstance(df, DataFrame)
    assert df.columns.to_list() == ["z", ("a1", "b1")]
    x = np.array([[1, 1, 1, 2, 2, 1, 1, 1], range(1, 9)]).T
    np.testing.assert_array_equal(df, x)


@pytest.mark.parametrize(
    ("by", "one", "two"),
    [
        ("x", [(7, 10)], [(11, 14)]),
        ("y", [(7, 8), (11, 12)], [(9, 10), (13, 14)]),
        ("z", [(7, 9), (12, 14)], [(10, 11)]),
    ],
)
def test_groupby(sf: SheetFrame, by, one, two):
    g = sf.groupby(by)
    assert len(g) == 2
    assert g[(1,)] == one
    assert g[(2,)] == two


@pytest.mark.parametrize(
    ("by", "v11", "v12", "v21", "v22"),
    [
        (["x", "y"], [(7, 8)], [(9, 10)], [(11, 12)], [(13, 14)]),
        (["x", "z"], [(7, 9)], [(10, 10)], [(12, 14)], [(11, 11)]),
        (["y", "z"], [(7, 8), (12, 12)], [(11, 11)], [(9, 9), (13, 14)], [(10, 10)]),
    ],
)
def test_groupby_list(sf: SheetFrame, by, v11, v12, v21, v22):
    g = sf.groupby(by)
    assert len(g) == 4
    assert g[(1, 1)] == v11
    assert g[(1, 2)] == v12
    assert g[(2, 1)] == v21
    assert g[(2, 2)] == v22



from functools import wraps
from itertools import cycle

import pandas as pd

# import spin.pandas as spd
from xlviews.axes import Axes, chart_position
from xlviews.config import rcParams
from xlviews.decorators import turn_off_screen_updating
from xlviews.style import palette
from xlviews.utils import format_label, label_func_from_list


def split_style(handle):
    """
    色指定などで、辞書を使うときの対応。
    color=('column', {'a': 'red', 'b': 'blue'})などを
    'column' と {('a',): 'red', 'b': 'blue'}に分割する。
    辞書のキーがタプルになっているのは、第一引数が最終的には
    リストになるため。
    color=('column', ['a', 'b', 'c'])の場合は、第2要素を順番に
    パレットから値を設定する。
    """
    if isinstance(handle, tuple):
        handle, style = handle
        if isinstance(style, dict):
            style = {
                key if isinstance(key, tuple) else (key,): value
                for key, value in style.items()
            }
        elif isinstance(style, list):
            style = [key if isinstance(key, tuple) else (key,) for key in style]
        return handle, style
    return handle, {}


def combine_handle(by, *handles, columns=None):
    """
    marker、colorなどをhandleをbyに追加する。
    なおかつ、リストでないものはリストにする。
    columnsを指定すると、marker, colorなどが直に形状や色を
    指定している場合を除外するために、columnsに含まれるものに限る。
    """

    def norm(x):
        if x is None:
            return []
        if not isinstance(x, list):
            return [x]
        return x[:]

    def none(x):
        return x if x else None

    by, *handles = [norm(x) for x in [by, *handles]]

    for handle in handles:
        for h in handle:
            if columns is not None and h not in columns:
                continue
            if h not in by:
                by.append(h)

    return [none(x) for x in [by, *handles]]


def get_label_name(label, by, key, **const_dict):
    """labelキーワード引数から凡例を作成する。"""
    if label and by:
        by_key = dict(zip(by, key, strict=False))
        for by_, key_ in const_dict.items():
            if by_ not in by:
                by_key[by_] = key_

        if isinstance(label, list):
            label = label_func_from_list(label)
        if callable(label):
            name = label(**by_key)
        else:
            name = label.format(**by_key)
    else:
        name = label
    if name == "auto":
        name = None
    return name


def autohandle(func):
    """
    マーカーや色のキーワード引数を処理する。
    この関数内で指定する任意のキーワード引数がhanldeになれる。
    これらのキーワード引数は、
      1. 'blue', 'o', など値を直に指定する。
      2. 'cad', 'pulse'などのようにカラム名で系列を分ける。
      3. ['cad', 'pulse']のように複数カラムで系列を分ける。
      4. ('cad', {'Y60': 'blue', 'Y70': 'red'})のように指定したスタイルで
          系列を分ける。
    """

    @wraps(func)
    def _func(self, *args, by=None, marker=None, color=None, alpha=None, **kwargs):
        marker, marker_style = split_style(marker)
        color, color_style = split_style(color)
        alpha, alpha_style = split_style(alpha)
        self.set_style(marker=marker_style, color=color_style, alpha=alpha_style)

        # if self.data.columns_level == 1:
        if self.data.columns_names is None:
            columns = self.data.columns
        else:
            columns = self.data.columns_names

        by, marker, color, alpha = combine_handle(
            by,
            marker,
            color,
            alpha,
            columns=columns,
        )
        self.set_handle(marker=marker, color=color, alpha=alpha)

        return func(self, *args, by=by, **kwargs)

    return _func


def autolegend(func):
    """
    凡例の処理を行う。
    """

    @wraps(func)
    def _func(self, *args, by=None, label=None, legend="auto", **kwargs):
        if legend == "auto":
            self.legend = dict(position=(1, 1))
        else:
            self.legend = legend
        if by and label == "auto":
            # label = '_'.join(['{' + x + '}' for x in by])
            label = label_func_from_list(by)

        axes = func(self, *args, by=by, label=label, **kwargs)

        if axes is None:
            return None

        if self.legend is None:
            axes.set_legend(None)
        elif isinstance(self.legend, dict):
            axes.set_legend(**self.legend)
        return axes

    return _func


class Element:
    def __init__(self, data):
        self.data = data
        self.sheet = None
        self.left = self.top = self.width = self.height = None
        self.handle = None
        self.const_dict = None
        self.style = None
        self.set_style()
        self.set_const_dict()
        self.axes = None
        self.chart = None
        self.series_collection = []
        self.by = None
        self.by_key = []
        self.legend = None
        self.x = None
        self.y = None
        self.chart_type = None
        self.include_in_layout = False

    def set_output(
        self,
        sheet=None,
        left=None,
        top=None,
        width=None,
        height=None,
        row=None,
        column=None,
        include_in_layout=False,
    ):
        self.sheet = sheet if sheet else self.data.sheet
        if column:
            self.left = self.sheet.range(1, column).left
        else:
            self.left = left
        if row:
            self.top = self.sheet.range(row, 1).top
        else:
            self.top = top
        self.left, self.top = chart_position(self.sheet, self.left, self.top)
        self.width, self.height = width, height
        self.include_in_layout = include_in_layout

    def set_style(self, **style):
        self.style = style.copy()

    def set_handle(self, **handle):
        self.handle = handle.copy()

    def set_const_dict(self, **const_dict):
        self.const_dict = const_dict.copy() if const_dict else {}
        # テールを持つシートフレームの場合は、そのデータを定数辞書に加える。
        if self.data.tail:
            self.const_dict.update(dict(self.data.tail.data.iloc[:, 0]))

    def set_axis(
        self,
        xticks="auto",
        yticks="auto",
        xlabel="auto",
        ylabel="auto",
        xformat="auto",
        yformat="auto",
        xscale=None,
        yscale=None,
    ):
        axes = self.axes
        if xscale:
            axes.set_xscale(xscale)
        if yscale:
            axes.set_yscale(yscale)

        x = self.x
        label_x = x[0] if isinstance(x, tuple) else x
        x = label_x.split("_")[0]
        y = self.y
        label_y = y[0] if isinstance(y, tuple) else y
        y = label_y.split("_")[0]

        if xticks == "auto":
            xticks = rcParams.get(f"axis.ticks.{x}", None)
        if yticks == "auto":
            yticks = rcParams.get(f"axis.ticks.{y}", None)
        if xticks:
            axes.set_xticks(*xticks)
        if yticks:
            axes.set_yticks(*yticks)

        if xformat == "auto":
            xformat = rcParams.get(f"axis.format.{x}", None)
        if yformat == "auto":
            yformat = rcParams.get(f"axis.format.{y}", None)
        axes.set_xtick_labels(format=xformat)
        axes.set_ytick_labels(format=yformat)

        if xlabel:
            if xlabel == "auto":
                xlabel = rcParams.get(f"axis.label.{x}", label_x)
                if "_" in label_x and "[" in xlabel:
                    xlabel = label_x + " " + xlabel[xlabel.index("[") :]
            axes.set_xlabel(xlabel)
        if ylabel:
            if ylabel == "auto":
                ylabel = rcParams.get(f"axis.label.{y}", label_y)
                if "_" in label_y and "[" in ylabel:
                    ylabel = label_y + " " + ylabel[ylabel.index("[") :]
            axes.set_ylabel(ylabel)

    @turn_off_screen_updating
    @autohandle
    @autolegend
    def _plot(
        self,
        x,
        y,
        chart_type=None,
        by=None,
        label=None,
        sel=None,
        sheet=None,
        left=None,
        top=None,
        width=None,
        height=None,
        row=None,
        column=None,
        axes=None,
        title="auto",
        line=False,
        line_width=False,
        name="{title_}_{x_}_{y_}",
        include_in_layout=False,
        **kwargs,
    ):
        """
        プロット本体

        Parameters
        ----------
        x, y : str
            シートフレームのカラム名
        chart_type : str
            エクセルチャートの種類
        by : str, list of str
            系列を分けるシートフレームのカラム名
        label : str
            ラベル文字列 '{cad}'などでシートフレームのカラムを含めることが
            できる。
        sel : dict
            シートフレームを選別する。
        sheet : xlwings.Sheet, optional
        left, top, width, height : int, optional
            チャートの位置と大きさ
        row, column : int, optional
            チャートのセル位置
        axes : xlviews.axes.Axes, optional
            重ね書きするときに指定する。
        title : str, optional
            タイトル
        line : bool, optional
            線を書くかどうか
        line_width : int, bool, optional
            線の太さ
        name : str
            チャートの名前
        data : xlviews.SheetFrame
            追加のデータをプロットするときのシートフレーム

        Returns
        -------
        xlviews.axes.Axes
        """
        if title == "auto" and hasattr(self.data, "title"):
            title = self.data.title

        if axes or self.axes:
            if axes:
                self.axes = axes
                self.chart = axes.chart
            is_new_axes = False
        else:
            self.set_output(
                sheet=sheet,
                left=left,
                top=top,
                width=width,
                height=height,
                row=row,
                column=column,
                include_in_layout=include_in_layout,
            )
            is_new_axes = True

        data = self.data
        self.x = x
        self.y = y
        self.chart_type = chart_type

        if isinstance(sel, dict):
            sel = data.select(**sel)
        else:
            sel = None
        grouped = data.groupby(by, sel)

        columns = data.columns
        if (
            data.columns_level == 1
            and (x not in columns or y not in columns)
            and not isinstance(x, tuple)
            and not isinstance(y, tuple)
        ):
            # Wide-format のデータフレーム。
            # 横方向にx軸、y軸をとる。
            by = data.index_columns
            if x in data.wide_columns and y in data.wide_columns:
                header = "both"
                index_x = data.index(x)
                index_y = data.index(y)
            elif x in data.wide_columns:
                header = "x"
                index = data.index(x)
            elif y in data.wide_columns:
                header = "y"
                index = data.index(y)
            else:
                raise ValueError("不明なカラム名", x, y)

            if sel is None:
                sel = [True] * len(data)
            if label == "auto":
                df = data[by][sel]
                df = spd.drop_const(df)
                label = list(df.columns)
                label = label_func_from_list(label)
            start = data.cell.row + 1
            column = data.cell.column
            for k in range(len(data)):
                row = start + k
                if sel[k]:
                    key = data.sheet.range(
                        (row, column),
                        (row, column + data.index_level),
                    ).value
                    series_name = get_label_name(label, by, key, **self.const_dict)
                    by_key = dict(zip(by, key, strict=False))
                    if header == "both":
                        # XValues
                        index = data.sheet.range((row, index_x[0]), (row, index_x[-1]))
                        # Values
                        columns = data.sheet.range(
                            (row, index_y[0]),
                            (row, index_y[-1]),
                        )
                    elif header == "x":
                        columns = [data.cell.row, row]
                    else:
                        columns = [row, data.cell.row]
                    self.add_series(index, columns, series_name, axis=0)
                    self.by_key.append(by_key)
        elif data.columns_names is None:
            # 通常のデータフレーム
            columns = data.index([x, y])
            for key, index in sorted(grouped.items()):
                if by is None:
                    by_key = None
                else:
                    by_key = dict(zip(by, key, strict=False))
                series_name = get_label_name(label, by, key, **self.const_dict)
                self.add_series(index, columns, series_name)
                self.by_key.append(by_key)
        else:
            # 波形など、unstackしたデータフレーム
            start = data.row + data.columns_level
            index = [start, start + len(data) - 1]

            # 軸項目はカラムの最後
            items = [column[-1] for column in data.columns]
            offset = data.column
            xycols = []
            for key, index_ in grouped.items():
                series_name = get_label_name(label, by, key, **self.const_dict)
                if by is None:
                    by_key = None
                else:
                    by_key = dict(zip(by, key, strict=False))

                # FIXME : もっと簡単にかけないか？
                for cols in index_:
                    xcol = ycol = None
                    for col in range(cols[0], cols[1] + 1):
                        item = items[col - offset]
                        if x == item:
                            if xcol is None:
                                xcol = col
                            else:
                                raise ValueError("重複したカラムあり。byを指定")
                        if y == item:
                            if ycol is None:
                                ycol = col
                            else:
                                raise ValueError("重複したカラムあり。byを指定")
                    if xcol and ycol:
                        xycols.extend([xcol - offset, ycol - offset])
                        self.add_series(index, [xcol, ycol], series_name)
                        self.by_key.append(by_key)
            # カラムがマルチインデックスの場合には、タイトル文字列用に
            # カラムインデックスからデータフレームを作成する。
            if title or name:
                data = pd.DataFrame(data.columns, columns=data.columns_names)
                data = data.iloc[xycols]
                sel = None

        if not self.axes:
            return None

        self.by = by
        self.apply_style(line=line, line_width=line_width)

        if is_new_axes:
            self.set_axis(**kwargs)
            if title:
                if isinstance(title, list):
                    title = label_func_from_list(title)
                if isinstance(title, str) or callable(title):
                    title = format_label(data, title, sel, default=self.const_dict)
                else:
                    name = None
                if title != "auto":
                    self.axes.set_title(title)
                else:
                    title = ""
            title = title if title else ""
            if name:
                name = name.replace("?", title)
                name = name.replace("{title_}", title)
                name = name.replace("{x_}", x[0] if isinstance(x, tuple) else x)
                name = name.replace("{y_}", y[0] if isinstance(y, tuple) else y)
                name = format_label(data, name, sel, default=self.const_dict)
                self.axes.chart.name = name
            self.axes.tight_layout()
            self.axes.set_style()
        return self.axes

    def add_axes(self):
        self.axes = Axes(
            sheet=self.sheet,
            left=self.left,
            top=self.top,
            width=self.width,
            height=self.height,
            include_in_layout=self.include_in_layout,
        )
        self.axes.set_chart_type(self.chart_type)
        self.chart = self.axes.chart

    def add_series(self, index, columns, name, axis=1):
        """
        シリーズを追加する。

        Parameters
        ----------
        index : list or xlwings.Range
            データインデックス
            See also: xlviews.utils.multirange
        columns : int or list or xlwings.Range
            intの場合、yの値のみ、listの場合(x, y)の値
        name : tuple or str
            tuple の場合、(row, col)
        axis : int
            データの方向

        Returns
        -------
        series : Series
            シリーズオブジェクト
        """
        if not self.axes:
            self.add_axes()
        sheet = self.data.sheet
        series = self.axes.add_series(
            index=index,
            columns=columns,
            name=name,
            axis=axis,
            sheet=sheet,
            chart_type=self.chart_type,
        )
        self.series_collection.append(series)
        return series

    @property
    def series_styles(self):
        """
        series_collectionに対応するスタイルのリストを返す。
        各要素はseriesに対応し、値は、{handle: style}の辞書である。
        """
        handles = self.handle.keys()
        styles = [self._series_styles(handle) for handle in handles]
        return [
            dict(zip(handles, style, strict=False))
            for style in zip(*styles, strict=False)
        ]

    def _series_styles(self, handle):
        """
        series_styleから呼ばれる内部関数で、handleごとのスタイルを生成する
        ジェネレータ
        """
        handles = self.handle[handle]
        if handles is None:
            yield from [None] * len(self.series_collection)
            return

        try:
            indexes = [self.by.index(handle) for handle in handles]
        except (ValueError, AttributeError):
            # hanldeに'red', 'cirlcle'などスタイルを直に描いている場合
            for _, handle in zip(self.series_collection, cycle(handles)):
                yield handle
            return

        defaults = palette(handle, len(self.series_collection))
        style = self.style[handle]
        if not isinstance(style, dict):
            style = dict(zip(style, palette(handle, len(style)), strict=False))
        for series, by_key, default in zip(
            self.series_collection,
            self.by_key,
            defaults,
            strict=False,
        ):
            key = tuple(by_key[self.by[index]] for index in indexes)
            if key not in style:
                style[key] = default
            else:
                defaults.insert(defaults.index(default), default)
            yield style[key]

    def apply_style(self, **kwargs):
        for series, style in zip(
            self.series_collection,
            self.series_styles,
            strict=False,
        ):
            set_series_style(series, **style, **kwargs)

    def plot(self, x, y, **kwargs):
        self._plot(x, y, chart_type="XYScatterLinesNoMarkers", **kwargs)

    def scatter(self, x, y, **kwargs):
        self._plot(x, y, chart_type="XYScatter", **kwargs)


class Plot(Element):
    def __init__(self, x, y, data=None, label="auto", **kwargs):
        super().__init__(data)
        self.plot(x, y, label=label, **kwargs)


class Scatter(Element):
    def __init__(self, x, y, data=None, marker="o", label="auto", **kwargs):
        super().__init__(data)
        self.scatter(x, y, marker=marker, label=label, **kwargs)


class Bar(Element):
    def __init__(self, x, y, data=None, label="auto", **kwargs):
        super().__init__(data)
        self.bar(x, y, label=label, **kwargs)

    @turn_off_screen_updating
    def bar(self, x, y, label="auto", stacked=False, **kwargs):
        if isinstance(x, list):
            xs, x = x, x[0]
        else:
            xs = None

        chart_type = "ColumnStacked" if stacked else "ColumnClustered"

        ys = y if isinstance(y, list) else [y]
        for y in ys:
            if label == "auto" and len(ys) > 1:
                label_ = y
            else:
                label_ = label
            self._plot(x, y, chart_type=chart_type, label=label_, **kwargs)

        if xs:
            sheet = self.data.sheet
            xvalues = sheet.range(
                self.data.range(xs[0])[0],
                self.data.range(xs[-1], -1)[-1],
            )
            self.series_collection[0].XValues = xvalues.api
            self.axes.set_xlabel(None)



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame


@pytest.fixture
def sf(sheet: Sheet):
    df = DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    return SheetFrame(3, 3, data=df, style=False, sheet=sheet)


@pytest.mark.parametrize("number_format", ["0", "0.00", "0.00%"])
def test_number_format(sf: SheetFrame, number_format: str):
    sf.set_number_format(number_format, autofit=True)
    assert sf.range("a").number_format == number_format
    assert sf.range("b").number_format == number_format


def test_number_format_kwargs(sf: SheetFrame):
    sf.set_number_format(autofit=False, a="0", b="0.0")
    assert sf.range("a").number_format == "0"
    assert sf.range("b").number_format == "0.0"


def test_number_format_dict(sf: SheetFrame):
    sf.set_number_format({r"[ab]": "0.00"}, autofit=True)
    assert sf.get_number_format("a") == "0.00"
    assert sf.get_number_format("b") == "0.00"


def test_style_gray(sf: SheetFrame):
    sf.add_wide_column("u", range(3), autofit=True, number_format="0", style=True)
    sf.set_style(gray=True)
    assert sf.sheet["F2"].api.Font.Bold
    assert sf.sheet["D3"].api.Font.Bold
    assert sf.sheet["C3"].api.Interior.Color == 15658734
    assert sf.sheet["H2"].api.Interior.Color == 15658734
    assert sf.sheet["H7"].api.Interior.Color != 15658734


@pytest.mark.parametrize(
    ("alignment", "value"),
    [("left", -4131), ("center", -4108), ("right", -4152)],
)
def test_alignment(sf: SheetFrame, alignment: str, value: int):
    sf.set_alignment(alignment)
    assert sf.cell.api.HorizontalAlignment == value


def test_adjacent_column_width(sf: SheetFrame):
    sf.set_adjacent_column_width(10)
    assert sf.sheet["F1"].column_width == 10


def test_child_frame(sf: SheetFrame):
    cell = sf.get_child_cell()
    assert cell.get_address() == "$G$3"

    cell = sf.get_adjacent_cell(offset=3)
    assert cell.get_address() == "$J$3"

    df = DataFrame({"x": [1, 2], "y": [5, 6], "z": [7, 8]})
    sf_child = SheetFrame(parent=sf, data=df, style=False)
    assert sf_child.cell.get_address() == "$G$3"

    assert sf_child.parent is sf
    assert sf.children[0] is sf_child

    cell = sf.get_adjacent_cell()
    assert cell.get_address() == "$L$3"


def test_head_frame(sf: SheetFrame):
    df = DataFrame({"x": [1, 2], "y": [5, 6], "z": [7, 8]})
    sf_tail = SheetFrame(head=sf, data=df, number_format="0.00")
    assert sf_tail.cell.get_address() == "$C$9"
    assert sf.tail is sf_tail
    assert sf_tail.head is sf


def test_active_sheet(sheet: Sheet):
    sheet.name = "active"
    sheet.activate()
    df = DataFrame({"x": [1, 2], "y": [5, 6], "z": [7, 8]})
    sf = SheetFrame(100, 10, data=df, style=False)
    assert sf.sheet.name == "active"
    assert sf.cell.get_address() == "$J$100"



import numpy as np
import pytest
from pandas import DataFrame, MultiIndex, Series
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame


@pytest.fixture(scope="module")
def df():
    df = DataFrame(
        {
            "c": [1, 2, 3, 4, 5],
            "d": [11, 12, 13, 14, 15],
            "e": [21, 22, 23, 24, 25],
            "f": [31, 32, 33, 34, 35],
        },
    )
    x = [("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
    df.columns = MultiIndex.from_tuples(x, names=["a", "b"])
    return df


def test_df(df: DataFrame):
    assert len(df) == 5
    assert df.shape == (5, 4)

    x = [("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
    assert df.columns.to_list() == x
    assert df.columns.names == ["a", "b"]
    assert isinstance(df.columns, MultiIndex)

    assert df.index.to_list() == list(range(5))
    assert df.index.names == [None]


@pytest.fixture(scope="module")
def sf(df: DataFrame, sheet_module: Sheet):
    return SheetFrame(20, 4, data=df, style=False, sheet=sheet_module)


def test_value(sf: SheetFrame):
    v = sf.cell.expand().options(ndim=2).value
    assert len(v) == 7
    assert v[0] == ["a", "a1", "a1", "a2", "a2"]
    assert v[1] == ["b", "b1", "b2", "b1", "b2"]
    assert v[2] == [0, 1, 11, 21, 31]
    assert v[-1] == [4, 5, 15, 25, 35]


def test_init(sf: SheetFrame, sheet_module: Sheet):
    assert sf.cell.get_address() == "$D$20"
    assert sf.row == 20
    assert sf.column == 4
    assert sf.sheet.name == sheet_module.name
    assert sf.index_level == 1
    assert sf.columns_level == 2
    assert sf.columns_names == ["a", "b"]


def test_set_data_from_sheet(sf: SheetFrame):
    sf.set_data_from_sheet(index_level=1, columns_level=2)
    assert sf.has_index is True
    x = [("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
    assert sf.value_columns == x
    assert sf.columns_names == ["a", "b"]


def test_len(sf: SheetFrame):
    assert len(sf) == 5


def test_columns(sf: SheetFrame):
    x = [("a", "b"), ("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
    assert sf.columns == x


def test_value_columns(sf: SheetFrame):
    c = [("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
    assert sf.value_columns == c


def test_index_columns(sf: SheetFrame):
    assert sf.index_columns == [("a", "b")]


def test_init_index_false(df: DataFrame, sheet: Sheet):
    sf = SheetFrame(2, 3, data=df, index=False, style=False, sheet=sheet)

    assert sf.index_level == 0
    c = [("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
    assert sf.columns == c


def test_contains(sf: SheetFrame):
    assert None not in sf
    assert ("a", "b") in sf
    assert ("a1", "b1") in sf
    assert "a1" not in sf


def test_iter(sf: SheetFrame):
    x = [("a", "b"), ("a1", "b1"), ("a1", "b2"), ("a2", "b1"), ("a2", "b2")]
    assert list(sf) == x


@pytest.mark.parametrize(
    ("column", "relative", "index"),
    [
        (("a", "b"), True, 1),
        (("a1", "b1"), True, 2),
        (("a2", "b2"), True, 5),
        (("a1", "b2"), False, 6),
        (("a2", "b1"), False, 7),
        ([("a1", "b2"), ("a2", "b1")], True, [3, 4]),
        ([("a", "b"), ("a2", "b2")], False, [4, 8]),
        ("a", True, 1),
        (["b", "a"], True, [2, 1]),
        ("a", False, 20),
        (["a", "b"], False, [20, 21]),
    ],
)
def test_index(sf: SheetFrame, column, relative, index):
    assert sf.index(column, relative=relative) == index


@pytest.mark.parametrize("column", ["a", "b"])
def test_index_row(sf: SheetFrame, column):
    r = sf.index(column, relative=False)
    assert sf.sheet.range(r, sf.column).value == column


@pytest.mark.parametrize(
    ("column", "relative", "index"),
    [
        ({"a": "a1"}, True, (2, 3)),
        ({"a": "a2"}, True, (4, 5)),
        ({"a": "a1"}, False, (5, 6)),
        ({"a": "a2"}, False, (7, 8)),
    ],
)
def test_index_dict(sf: SheetFrame, column, relative, index):
    assert sf.index_dict(column, relative=relative) == index
    assert sf.index(column, relative=relative) == index


def test_data(sf: SheetFrame, df: DataFrame):
    df_ = sf.data
    np.testing.assert_array_equal(df_.index, df.index)
    np.testing.assert_array_equal(df_.index.names, df.index.names)
    np.testing.assert_array_equal(df_.columns, df.columns)
    np.testing.assert_array_equal(df_.columns.names, df.columns.names)
    np.testing.assert_array_equal(df_, df)
    assert df_.index.name == df.index.name
    assert df_.columns.name == df.columns.name


def test_range_all(sf: SheetFrame):
    assert sf.range_all().get_address() == "$D$20:$H$26"
    assert sf.range().get_address() == "$D$20:$H$26"


@pytest.mark.parametrize(
    ("start", "end", "address"),
    [
        (0, None, "$D$22"),
        (50, None, "$D$50"),
        (50, 100, "$D$50:$D$100"),
    ],
)
def test_range_index(sf: SheetFrame, start, end, address):
    assert sf.range_index(start, end).get_address() == address
    assert sf.range("index", start, end).get_address() == address


@pytest.mark.parametrize("start", [None, -1, False])
def test_range_index_error(sf: SheetFrame, start):
    with pytest.raises(ValueError, match="index start must be a specific row"):
        sf.range_index(start)

    with pytest.raises(ValueError, match="index start must be a specific row"):
        sf.range("index", start)


@pytest.mark.parametrize(
    ("column", "start", "end", "address"),
    [
        (("a2", "b2"), False, None, "$H$20:$H$26"),
        (("a1", "b1"), -1, None, "$E$20:$E$21"),
        (("a1", "b2"), 0, None, "$F$22"),
        (("a2", "b1"), None, None, "$G$22:$G$26"),
        (("a1", "b1"), 30, None, "$E$30"),
        (("a1", "b2"), 30, 40, "$F$30:$F$40"),
    ],
)
def test_range_column(sf: SheetFrame, column, start, end, address):
    assert sf.range_column(column, start, end).get_address() == address
    assert sf.range(column, start, end).get_address() == address


def test_getitem_tuple(sf: SheetFrame):
    s = sf[("a1", "b1")]
    assert isinstance(s, Series)
    assert s.name == ("a1", "b1")
    np.testing.assert_array_equal(s, [1, 2, 3, 4, 5])


def test_getitem_list(sf: SheetFrame):
    df = sf[[("a1", "b1"), ("a2", "b2")]]
    assert isinstance(df, DataFrame)
    assert df.columns.to_list() == [("a1", "b1"), ("a2", "b2")]
    x = np.array([[1, 2, 3, 4, 5], [31, 32, 33, 34, 35]]).T
    np.testing.assert_array_equal(df, x)


@pytest.mark.parametrize(
    ("a", "b", "sel"),
    [
        ("a1", None, [True, True, False, False]),
        ("a2", None, [False, False, True, True]),
        ("a1", "b1", [True, False, False, False]),
        ("a2", "b2", [False, False, False, True]),
        (["a1", "a2"], None, [True, True, True, True]),
        (["a1", "a2"], ["b2"], [False, True, False, True]),
        (("a1", "a2"), ("b1", "b1"), [True, False, True, False]),
    ],
)
def test_select(sf: SheetFrame, a, b, sel):
    if b is None:
        np.testing.assert_array_equal(sf.select(a=a), sel)
    else:
        np.testing.assert_array_equal(sf.select(a=a, b=b), sel)


@pytest.mark.parametrize(
    ("by", "one", "two"),
    [
        ("a", [(5, 6)], [(7, 8)]),
        ("b", [(5, 5), (7, 7)], [(6, 6), (8, 8)]),
    ],
)
def test_groupby(sf: SheetFrame, by, one, two):
    g = sf.groupby(by)
    assert len(g) == 2
    assert g[(f"{by}1",)] == one
    assert g[(f"{by}2",)] == two


def test_groupby_list(sf: SheetFrame):
    g = sf.groupby(["a", "b"])
    assert len(g) == 4
    assert g[("a1", "b1")] == [(5, 5)]
    assert g[("a1", "b2")] == [(6, 6)]
    assert g[("a2", "b1")] == [(7, 7)]
    assert g[("a2", "b2")] == [(8, 8)]


def test_groupby_none(sf: SheetFrame):
    g = sf.groupby(None)
    assert len(g) == 1
    assert g[()] == [(5, 8)]



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar, overload

import numpy as np
from pandas import DataFrame, Series

from xlviews.utils import iter_columns

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence

    from xlwings import Range

    from xlviews.range import RangeCollection
    from xlviews.sheetframe import SheetFrame

H = TypeVar("H")
T = TypeVar("T")


def _to_dict(keys: Iterable[H], values: Iterable[T]) -> dict[H, list[T]]:
    result = {}

    for key, value in zip(keys, values, strict=True):
        result.setdefault(key, []).append(value)

    return result


def create_group_index(
    a: Sequence | Series | DataFrame,
) -> dict[tuple, list[tuple[int, int]]]:
    df = a.reset_index(drop=True) if isinstance(a, DataFrame) else DataFrame(a)

    dup = df[df.ne(df.shift()).any(axis=1)]

    start = dup.index.to_numpy()
    end = np.r_[start[1:] - 1, len(df) - 1]

    keys = [tuple(v) for v in dup.to_numpy()]
    values = [(int(s), int(e)) for s, e in zip(start, end, strict=True)]

    return _to_dict(keys, values)


class Grouper:
    sf: SheetFrame
    by: list[str]
    grouped: dict[tuple, list[tuple[int, int]]]

    def __init__(self, sf: SheetFrame, by: str | list[str] | None = None) -> None:
        self.sf = sf
        self.by = list(iter_columns(sf, by)) if by else []
        self.grouped = sf.groupby(self.by)

    def __len__(self) -> int:
        return len(self.grouped)

    def keys(self) -> Iterator[tuple]:
        yield from self.grouped.keys()

    def values(self) -> Iterator[list[tuple[int, int]]]:
        yield from self.grouped.values()

    def items(self) -> Iterator[tuple[tuple, list[tuple[int, int]]]]:
        yield from self.grouped.items()

    def __iter__(self) -> Iterator[tuple]:
        yield from self.keys()

    def __getitem__(self, key: tuple) -> list[tuple[int, int]]:
        return self.grouped[key]

    def range(self, column: str, key: tuple) -> RangeCollection:
        return self.sf.range(column, self[key])

    def first_range(self, column: str, key: tuple) -> Range:
        return self.sf.range(column, self[key][0][0])

    @overload
    def ranges(self, column: str) -> Iterator[RangeCollection]: ...

    @overload
    def ranges(self, column: tuple) -> Iterator[Range]: ...

    @overload
    def ranges(self, column: None = None, **kwargs) -> Iterator[Iterator[Range]]: ...

    def ranges(
        self,
        column: str | tuple | None = None,
        **kwargs,
    ) -> Iterator[RangeCollection | Range | Iterator[Range]]:
        if isinstance(column, str):
            for key in self:
                yield self.range(column, key)
        elif column is None:
            for column in self:
                yield self.ranges(column, **kwargs)
        else:
            kwargs.update(dict(zip(self.by, column, strict=True)))
            yield from self.sf.ranges(**kwargs)

    def first_ranges(self, column: str) -> Iterator[Range]:
        for key in self:
            yield self.first_range(column, key)



from __future__ import annotations

import re
import warnings
from collections import OrderedDict
from typing import TYPE_CHECKING

import matplotlib as mpl
import xlwings as xw
from xlwings import Range
from xlwings.constants import DVType, FormatConditionOperator

from xlviews.config import rcParams

if TYPE_CHECKING:
    from collections.abc import Iterator

    from pandas import DataFrame

    from xlviews.sheetframe import SheetFrame


def constant(type_: str, name: str | None = None) -> int:
    """Return the Excel constant.

    Args:
        type_ (str): The type name.
        name (str): The name.

    Examples:
        >>> constant("BordersIndex", "EdgeTop")
        8
    """
    if name is None:
        if "." in type_:
            type_, name = type_.split(".")
        else:
            type_, name = "Constants", type_

    if not name.startswith("xl"):
        name = "xl" + name[0].upper() + name[1:]

    type_ = getattr(xw.constants, type_)

    return getattr(type_, name)


def int_to_column_name(n: int) -> str:
    """Return the Excel column name from an integer.

    Examples:
        >>> int_to_column_name(1)
        'A'
        >>> int_to_column_name(26)
        'Z'
        >>> int_to_column_name(27)
        'AA'
    """
    name = ""

    while n > 0:
        n -= 1
        name = chr(n % 26 + 65) + name
        n //= 26

    return name


def rgb(
    color: int | tuple[int, int, int] | str,
    green: int | None = None,
    blue: int | None = None,
) -> int:
    """Return the RGB color integer.

    Args:
        color (int, tuple[int, int, int], or str): The color or red value.
        green (int): The green value.
        blue (int): The blue value.

    Examples:
        >>> rgb(4)
        4

        >>> rgb((100, 200, 40))
        2672740

        >>> rgb("pink")
        13353215

        >>> rgb("#123456")
        5649426
    """
    if isinstance(color, int) and green is None and blue is None:
        return color

    if all(isinstance(x, int) for x in [color, green, blue]):
        return color + green * 256 + blue * 256 * 256  # type: ignore

    if isinstance(color, str):
        color = mpl.colors.cnames.get(color, color)  # type: ignore

        if not isinstance(color, str) or not color.startswith("#") or len(color) != 7:
            raise ValueError("Invalid color format. Expected #xxxxxx.")

        return rgb(int(color[1:3], 16), int(color[3:5], 16), int(color[5:7], 16))

    if isinstance(color, tuple):
        return rgb(*color)

    raise ValueError("Invalid color format. Expected #xxxxxx.")


def iter_columns(sf: DataFrame | SheetFrame, columns: str | list[str]) -> Iterator[str]:
    """Yield the columns in the order of appearance with colon notation.

    Examples:
        >>> from pandas import DataFrame
        >>> df = DataFrame([[1, 2, 3]], columns=["A", "B", "C"])
        >>> list(iter_columns(df, "B"))
        ['B']
        >>> list(iter_columns(df, ":B"))
        ['A', 'B']
        >>> list(iter_columns(df, "::B"))
        ['A']
    """
    if isinstance(columns, str):
        columns = [columns]

    cs = [c for c in sf if isinstance(c, str)]

    for c in columns:
        if c.startswith("::"):
            yield from cs[: cs.index(c[2:])]
        elif c.startswith(":"):
            yield from cs[: cs.index(c[1:]) + 1]
        else:
            yield c


def add_validate_list(
    rng: Range,
    value: list[object],
    default: object | None = None,
) -> None:
    if default:
        rng.value = default

    type_ = DVType.xlValidateList
    operator = FormatConditionOperator.xlEqual
    formula = ",".join(map(str, value))

    rng.api.Validation.Add(Type=type_, Operator=operator, Formula1=formula)


def label_func_from_list(columns, post=None):
    """
    カラム名のリストからラベル関数を作成して返す。

    Parameters
    ----------
    columns : list of str
        カラム名のリスト
    post : str, optional
        追加文字列

    Returns
    -------
    callable
    """

    def get_format(t):
        name_ = f"column.label.{t}"
        if name_ in rcParams:
            return rcParams[name_]
        return "{" + t + "}"

    fmt_dict = OrderedDict()
    for column in columns:
        fmt_dict[column] = get_format(column)

    def func(**by_key):
        labels = []
        for by, fmt in fmt_dict.items():
            key = by_key[by]
            if isinstance(fmt, str):
                label = fmt.format(**{by: key})
            else:
                label = fmt(key)
            labels.append(label)
        return "_".join(labels) + ("_" + post if post else "")

    return func


def format_label(data, fmt, sel=None, default=None):
    dict_ = default.copy() if default else {}
    if callable(fmt):
        for column in data.columns:
            try:
                values = data[column]
            except TypeError:
                continue
            if sel is not None:
                values = values[sel]
            values = values.unique()
            if len(values) == 1:
                dict_[column] = values[0]
        return fmt(**dict_)
    keys = re.findall(r"{([\w.]+)(?:}|:)", fmt)
    for column in keys:
        if column in data.columns:
            values = data[column]
            if sel is not None:
                values = values[sel]
            values = values.unique()
            if len(values) == 1:
                dict_[column] = values[0]
    for key in keys:
        if key not in dict_:
            warnings.warn(
                f"タイトル文字列に含まれる'{key}'が、"
                "dfに含まれないか、単一ではない。",
            )
            dict_[key] = "XXX"
    return fmt.format(**dict_)


# def get_sheet(book, name):
#     try:
#         return book.sheets[name]
#     except Exception:
#         return book.sheets.add(name, after=book.sheets(book.sheets.count))

# def get_range(book, name, title=False):
#     for sheet in book.sheets:
#         try:
#             range_ = sheet.names(name).refers_to_range
#             if title:
#                 start = range_[0, 0].offset(-1, 0)
#                 if start.value:
#                     return sheet.range(start, range_[-1, -1])
#             else:
#                 return range_
#         except Exception:
#             continue


# def copy_range(book_from, sheet_to, name, title=False):
#     range_ = get_range(book_from, name.replace("-", "__"), title=title)
#     range_.api.CopyPicture()  # Appearance:=xlScreen, Format:=xlPicture)
#     # sheet_to.activate()
#     # sheet_to.range('A1').api.Select()
#     sheet_to.api.Paste()
#     sheet_to.pictures[-1].name = name.replace("__", "-")

# def get_chart(book, name):
#     for sheet in book.sheets:
#         try:
#             return sheet.charts(name)
#         except Exception:
#             continue


# def copy_chart(book_from, sheet_to, name):
#     chart = get_chart(book_from, name)
#     # chart.api[1].ChartArea.Copy()
#     chart.api[0].Copy()
#     # sheet_to.api.Paste()
#     # sheet_to.activate()
#     # sheet_to.range('A1').api.Select()
#     sheet_to.api.PasteSpecial(Format="図 (PNG)", Link=False, DisplayAsIcon=False)
#     sheet_to.pictures[-1].name = name

# def outline_group(sheet, start: int, end: int, axis=0):
#     """
#     セルをグループする。
#     """
#     outline = sheet.api.Outline
#     if axis == 0:
#         outline.SummaryRow = constant("SummaryRow.xlSummaryAbove")
#         sheet.range((start, 1), (end, 1)).api.EntireRow.Group()
#     else:
#         outline.SummaryColumn = constant("SummaryColumn.xlSummaryOnLeft")
#         sheet.range((1, start), (1, end)).api.EntireRow.Group()


# def show_group(start: int, axis=0, show=True):
#     app = xw.apps.active
#     if axis == 0:
#         app.api.ExecuteExcel4Macro(f"SHOW.DETAIL(1,{start},{show})")
#     else:
#         raise ValueError("未実装")


# def hide_group(start: int, axis=0):
#     show_group(start, axis=axis, show=False)


# def outline_levels(sheet, levels: int, axis=0):
#     if axis == 0:
#         sheet.api.Outline.ShowLevels(RowLevels=levels)
#     else:
#         sheet.api.Outline.ShowLevels(ColumnLevels=levels)



import numpy as np
import pytest
from pandas import DataFrame, Series

from xlviews.grouper import _to_dict


@pytest.mark.parametrize(
    ("keys", "values", "expected"),
    [
        ([1, 2, 1], ["a", "b", "c"], {1: ["a", "c"], 2: ["b"]}),
        (["x", "y", "x"], [10, 20, 30], {"x": [10, 30], "y": [20]}),
        ([True, False, True], [1.1, 2.2, 3.3], {True: [1.1, 3.3], False: [2.2]}),
        ([], [], {}),
        ([None, None], ["a", "b"], {None: ["a", "b"]}),
    ],
)
def test_to_dict(keys, values, expected):
    assert _to_dict(keys, values) == expected


@pytest.mark.parametrize("func", [lambda x: x, np.array, Series])
def test_create_group_index_series(func):
    from xlviews.grouper import create_group_index

    values = [1, 1, 2, 2, 2, 3, 3, 1, 1, 2, 2, 3, 3]
    index = create_group_index(func(values))
    assert index[(1,)] == [(0, 1), (7, 8)]
    assert index[(2,)] == [(2, 4), (9, 10)]
    assert index[(3,)] == [(5, 6), (11, 12)]


@pytest.mark.parametrize("func", [lambda x: x, DataFrame])
def test_create_group_index_dataframe(func):
    from xlviews.grouper import create_group_index

    values = [[1, 2], [1, 2], [3, 4], [3, 4], [1, 2], [3, 4], [3, 4]]
    index = create_group_index(func(values))
    assert index[(1, 2)] == [(0, 1), (4, 4)]
    assert index[(3, 4)] == [(2, 3), (5, 6)]



import string

import numpy as np
import pytest
from pandas import DataFrame, MultiIndex
from xlwings import Sheet

from xlviews.grouper import Grouper
from xlviews.sheetframe import SheetFrame


@pytest.fixture(scope="module")
def sf(sheet_module: Sheet):
    a = ["a"] * 8 + ["b"] * 8
    b = (["c"] * 4 + ["d"] * 4) * 2
    c = np.repeat(range(1, 9), 2)
    d = ["x", "y"] * 8
    df = DataFrame(np.arange(16 * 6).reshape(16, 6).T)
    df.columns = MultiIndex.from_arrays([a, b, c, d], names=["s", "t", "r", "i"])
    return SheetFrame(2, 2, data=df, index=True, style=False, sheet=sheet_module)


@pytest.mark.parametrize(
    ("by", "n"),
    [
        ("s", 2),
        ("t", 2),
        ("r", 8),
        ("i", 2),
        (["s", "t"], 4),
        (["s", "t", "i"], 8),
        (["t", "i"], 4),
        (["s", "t", "r", "i"], 16),
    ],
)
def test_len(sf: SheetFrame, by, n: int):
    gr = Grouper(sf, by)
    assert len(gr) == n


@pytest.fixture(scope="module")
def gr(sf: SheetFrame):
    return Grouper(sf, ["s", "t"])


def test_keys(gr: Grouper):
    keys = [("a", "c"), ("a", "d"), ("b", "c"), ("b", "d")]
    assert list(gr.keys()) == keys


def test_values(gr: Grouper):
    values = [[(3, 6)], [(7, 10)], [(11, 14)], [(15, 18)]]
    assert list(gr.values()) == values


@pytest.mark.parametrize(
    ("column", "value"),
    [
        (("a", "c"), [2, 3, 4, 5]),
        (("a", "d"), [6, 7, 8, 9]),
        (("b", "c"), [10, 11, 12, 13]),
        (("b", "d"), [14, 15, 16, 17]),
    ],
)
def test_ranges(gr: Grouper, column: tuple, value):
    for rng, i in zip(gr.ranges(column), value, strict=True):
        c = string.ascii_uppercase[i]
        assert rng.get_address() == f"${c}$6:${c}$11"


def test_ranges_none(gr: Grouper):
    values = [[2, 3, 4, 5], [6, 7, 8, 9], [10, 11, 12, 13], [14, 15, 16, 17]]
    for it, value in zip(gr.ranges(), values, strict=True):
        for rng, i in zip(it, value, strict=True):
            c = string.ascii_uppercase[i]
            assert rng.get_address() == f"${c}$6:${c}$11"


def test_ranges_kwargs(gr: Grouper):
    values = [[2, 4], [6, 8], [10, 12], [14, 16]]
    for it, value in zip(gr.ranges(i="x"), values, strict=True):
        for rng, i in zip(it, value, strict=True):
            c = string.ascii_uppercase[i]
            assert rng.get_address() == f"${c}$6:${c}$11"



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame
from xlviews.utils import rgb


@pytest.fixture
def sf(df: DataFrame, sheet: Sheet):
    return SheetFrame(3, 3, data=df, table=True, sheet=sheet)


@pytest.mark.parametrize(
    ("func", "color"),
    [
        ("count", "gray"),
        ("max", "#FF7777"),
        ("mean", "#33aa33"),
        ("min", "#7777FF"),
        ("soa", "#5555FF"),
        ("std", "#aaaaaa"),
        ("sum", "purple"),
    ],
)
def test_value_style(sf: SheetFrame, func, color):
    sf = sf.statsframe(func, by="x", table=True)
    sf.set_value_style("func")
    for c in ["a", "b", "c"]:
        rng = sf.range(c)
        assert rgb(rng.font.color) == rgb(color)
        if func in ["soa", "sum"]:
            assert rng.font.italic
        else:
            assert not rng.font.italic
        if func == "soa":
            assert rng.number_format == "0.0%"



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.grouper import Grouper
from xlviews.sheetframe import SheetFrame


@pytest.fixture(scope="module")
def sf(sheet_module: Sheet):
    a = ["c"] * 10
    b = ["s"] * 5 + ["t"] * 5
    c = ([100] * 2 + [200] * 3) * 2
    x = list(range(10))
    y = list(range(10, 20))
    df = DataFrame({"a": a, "b": b, "c": c, "x": x, "y": y})
    df = df.set_index(["a", "b", "c"])
    return SheetFrame(2, 2, data=df, index=True, style=False, sheet=sheet_module)


@pytest.mark.parametrize(
    ("by", "n"),
    [
        ("a", 1),
        ("b", 2),
        ("c", 2),
        (["a", "b"], 2),
        (["a", "c"], 2),
        (["b", "c"], 4),
        (["a", "b", "c"], 4),
    ],
)
def test_len(sf: SheetFrame, by, n: int):
    gr = Grouper(sf, by)
    assert len(gr) == n


@pytest.fixture(scope="module")
def gr(sf: SheetFrame):
    return Grouper(sf, ["a", "c"])


def test_keys(gr: Grouper):
    keys = [("c", 100), ("c", 200)]
    assert list(gr.keys()) == keys


def test_values(gr: Grouper):
    values = [[(3, 4), (8, 9)], [(5, 7), (10, 12)]]
    assert list(gr.values()) == values


def test_items(gr: Grouper):
    assert next(gr.items()) == (("c", 100), [(3, 4), (8, 9)])


def test_iter(gr: Grouper):
    assert next(iter(gr)) == ("c", 100)


@pytest.mark.parametrize(
    ("key", "value"),
    [
        (("c", 100), [(3, 4), (8, 9)]),
        (("c", 200), [(5, 7), (10, 12)]),
    ],
)
def test_getitem(gr: Grouper, key, value):
    assert gr[key] == value


@pytest.mark.parametrize(
    ("key", "value"),
    [
        (("c", 100), "$E$3:$E$4,$E$8:$E$9"),
        (("c", 200), "$E$5:$E$7,$E$10:$E$12"),
    ],
)
def test_range(gr: Grouper, key, value):
    assert gr.range("x", key).get_address() == value


@pytest.mark.parametrize(
    ("key", "value"),
    [
        (("c", 100), "$B$3"),
        (("c", 200), "$B$5"),
    ],
)
def test_first_range(gr: Grouper, key, value):
    assert gr.first_range("a", key).get_address() == value



import numpy as np
import pytest
from scipy.stats import norm

from xlviews.distframe import DistFrame
from xlviews.sheetframe import SheetFrame


def test_init_data(sf: SheetFrame):
    from xlviews.distframe import get_init_data

    df = get_init_data(sf, ["a", "b"], ["x", "y"])
    c = ["a_n", "a_v", "a_s", "b_n", "b_v", "b_s"]
    assert df.columns.to_list() == c
    assert df.index.names == ["x", "y"]
    assert len(df) == 14


def test_dist_func_str():
    from xlviews.distframe import get_dist_func

    df = get_dist_func("norm", ["a", "b"])
    assert df == {"a": "norm", "b": "norm"}


def test_dist_func_dict():
    from xlviews.distframe import get_dist_func

    df = get_dist_func({"a": "none"}, ["a", "b"])
    assert df == {"a": "none", "b": "norm"}


@pytest.fixture(scope="module")
def sfd(sf: SheetFrame):
    from xlviews.distframe import DistFrame

    return DistFrame(sf, ["a", "b"], by=["x", "y"])


@pytest.mark.parametrize(
    ("cell", "value"),
    [
        ("G4", 1),
        ("I4", 1),
        ("J4", 1),
        ("I7", 4),
        ("J7", 4),
        ("I17", 2),
        ("J17", 2),
        ("K4", norm.ppf(1 / 6)),
        ("N5", norm.ppf(2 / 6)),
        ("K6", norm.ppf(3 / 6)),
        ("N7", norm.ppf(4 / 6)),
        ("K8", norm.ppf(5 / 6)),
        ("N16", norm.ppf(1 / 3)),
        ("K17", norm.ppf(2 / 3)),
    ],
)
def test_distframe(sfd: DistFrame, cell: str, value: float):
    v = sfd.sheet[cell].value
    assert v is not None
    np.testing.assert_allclose(v, value)



import string

import numpy as np
import pytest
from pandas import DataFrame, MultiIndex, Series
from xlwings import Sheet

from xlviews.sheetframe import SheetFrame


@pytest.fixture(scope="module")
def sf(sheet_module: Sheet):
    a = ["a"] * 8 + ["b"] * 8
    b = (["c"] * 4 + ["d"] * 4) * 2
    c = np.repeat(range(1, 9), 2)
    d = ["x", "y"] * 8
    df = DataFrame(np.arange(16 * 6).reshape(16, 6).T)
    df.columns = MultiIndex.from_arrays([a, b, c, d], names=["s", "t", "r", "i"])
    return SheetFrame(2, 2, data=df, index=True, style=False, sheet=sheet_module)


def test_value(sf: SheetFrame):
    v = sf.cell.expand().options(ndim=2).value
    assert len(v) == 10
    assert v[2][:5] == ["r", 1, 1, 2, 2]
    assert v[3][:5] == ["i", "x", "y", "x", "y"]
    assert v[-1][:5] == [5, 5, 11, 17, 23]


def test_init(sf: SheetFrame, sheet_module: Sheet):
    assert sf.cell.get_address() == "$B$2"
    assert sf.row == 2
    assert sf.column == 2
    assert sf.sheet.name == sheet_module.name
    assert sf.index_level == 1
    assert sf.columns_level == 4
    assert sf.columns_names == ["s", "t", "r", "i"]


def test_len(sf: SheetFrame):
    assert len(sf) == 6


def test_columns(sf: SheetFrame):
    columns = sf.columns
    assert columns[0] == ("s", "t", "r", "i")
    assert columns[1] == ("a", "c", 1, "x")
    assert columns[-1] == ("b", "d", 8, "y")


def test_value_columns(sf: SheetFrame):
    columns = sf.value_columns
    assert columns[0] == ("a", "c", 1, "x")
    assert columns[-1] == ("b", "d", 8, "y")


def test_index_columns(sf: SheetFrame):
    assert sf.index_columns == [("s", "t", "r", "i")]


@pytest.mark.parametrize(
    ("column", "index"),
    [
        (("s", "t", "r", "i"), 2),
        (("a", "d", 3, "x"), 7),
        ([("b", "c", 6, "y"), ("b", "d", 8, "x")], [14, 17]),
        ("s", 2),
        (["t", "i"], [3, 5]),
    ],
)
def test_index(sf: SheetFrame, column, index):
    assert sf.index(column) == index


@pytest.mark.parametrize("column", ["s", "t", "r", "i"])
def test_index_row(sf: SheetFrame, column):
    r = sf.index(column)
    assert sf.sheet.range(r, sf.column).value == column


@pytest.mark.parametrize(
    ("column", "index"),
    [
        ({"s": "a"}, (3, 10)),
        ({"s": "b", "t": "c"}, (11, 14)),
        ({"r": 5}, (11, 12)),
        ({"r": 7, "i": "y"}, (16, 16)),
    ],
)
def test_index_dict(sf: SheetFrame, column, index):
    assert sf.index_dict(column) == index
    assert sf.index(column) == index


def test_index_dict_error(sf: SheetFrame):
    with pytest.raises(NotImplementedError):
        sf.index_dict({"t": "c"})


def test_range_all(sf: SheetFrame):
    assert sf.range_all().get_address() == "$B$2:$R$11"
    assert sf.range().get_address() == "$B$2:$R$11"


@pytest.mark.parametrize(
    ("start", "end", "address"),
    [
        (0, None, "$B$6"),
        (50, None, "$B$50"),
        (15, 16, "$B$15:$B$16"),
    ],
)
def test_range_index(sf: SheetFrame, start: int | None, end, address):
    assert sf.range_index(start, end).get_address() == address
    assert sf.range("index", start, end).get_address() == address


@pytest.mark.parametrize(
    ("column", "start", "end", "address"),
    [
        (("a", "c", 1, "y"), False, None, "$D$2:$D$11"),
        (("a", "d", 4, "y"), -1, None, "$J$2:$J$5"),
        (("a", "d", 3, "x"), 0, None, "$G$6"),
        (("b", "d", 7, "x"), None, None, "$O$6:$O$11"),
        (("a", "d", 3, "y"), 100, None, "$H$100"),
    ],
)
def test_range_column(sf: SheetFrame, column, start, end, address):
    assert sf.range_column(column, start, end).get_address() == address
    assert sf.range(column, start, end).get_address() == address


@pytest.mark.parametrize(
    ("column", "start", "end", "address"),
    [
        (("a", "d", 3, "y"), [(3, 5), (7, 8)], None, "$H$3:$H$5,$H$7:$H$8"),
    ],
)
def test_range_start_list(sf: SheetFrame, column, start, end, address):
    assert sf.range(column, start, end).get_address() == address


def test_range_column_error(sf: SheetFrame):
    with pytest.raises(NotImplementedError):
        sf.range_column("s")

    with pytest.raises(NotImplementedError):
        sf.range("t")


def test_getitem_tuple(sf: SheetFrame):
    s = sf[("a", "d", 4, "x")]
    assert isinstance(s, Series)
    assert s.name == ("a", "d", 4, "x")
    np.testing.assert_array_equal(s, [36, 37, 38, 39, 40, 41])
    s = sf["a", "d", 4, "x"]
    np.testing.assert_array_equal(s, [36, 37, 38, 39, 40, 41])


def test_getitem_list(sf: SheetFrame):
    df = sf[[("a", "d", 4, "x"), ("a", "d", 4, "y")]]
    assert isinstance(df, DataFrame)
    assert df.columns.to_list() == [("a", "d", 4, "x"), ("a", "d", 4, "y")]
    x = np.arange(36, 48).reshape(2, 6).T
    np.testing.assert_array_equal(df, x)


@pytest.mark.parametrize(
    ("kwargs", "sel"),
    [
        ({"s": "a"}, [True] * 8 + [False] * 8),
        ({"i": "x"}, [True, False] * 8),
        ({"s": "a", "i": "x"}, [True, False] * 4 + [False] * 8),
        ({"r": (3, 6)}, [False] * 4 + [True] * 8 + [False] * 4),
        ({"r": [1, 8]}, [True] * 2 + [False] * 12 + [True] * 2),
    ],
)
def test_select(sf: SheetFrame, kwargs, sel):
    x = sf.select(**kwargs)
    np.testing.assert_array_equal(x, sel)


def test_ranges(sf: SheetFrame):
    for rng, i in zip(sf.ranges(), range(2, 18), strict=True):
        c = string.ascii_uppercase[i]
        assert rng.get_address() == f"${c}$6:${c}$11"


def test_ranges_sel(sf: SheetFrame):
    sel = [False] * 8 + [True] * 8
    for rng, i in zip(sf.ranges(sel), range(10, 18), strict=True):
        c = string.ascii_uppercase[i]
        assert rng.get_address() == f"${c}$6:${c}$11"


def test_ranges_kwargs(sf: SheetFrame):
    for rng, i in zip(sf.ranges(i="y"), range(3, 18, 2), strict=True):
        c = string.ascii_uppercase[i]
        assert rng.get_address() == f"${c}$6:${c}$11"


@pytest.mark.parametrize(
    ("by", "result"),
    [
        ("s", {("a",): [(3, 10)], ("b",): [(11, 18)]}),
        (
            ["s", "t"],
            {
                ("a", "c"): [(3, 6)],
                ("a", "d"): [(7, 10)],
                ("b", "c"): [(11, 14)],
                ("b", "d"): [(15, 18)],
            },
        ),
        (
            ["s", "i"],
            {
                ("a", "x"): [(3, 3), (5, 5), (7, 7), (9, 9)],
                ("a", "y"): [(4, 4), (6, 6), (8, 8), (10, 10)],
                ("b", "x"): [(11, 11), (13, 13), (15, 15), (17, 17)],
                ("b", "y"): [(12, 12), (14, 14), (16, 16), (18, 18)],
            },
        ),
        (None, {(): [(3, 18)]}),
    ],
)
def test_groupby(sf: SheetFrame, by, result):
    g = sf.groupby(by)
    assert len(g) == len(result)
    for k, v in g.items():
        assert result[k] == v



import pytest
from xlwings import Sheet

from xlviews.range import RangeCollection


def test_reference_str(sheet_module: Sheet):
    from xlviews.range import reference

    assert reference("x", sheet_module) == "x"


def test_reference_range(sheet_module: Sheet):
    from xlviews.range import reference

    cell = sheet_module.range(4, 5)

    ref = reference(cell)
    assert ref == f"={sheet_module.name}!$E$4"


def test_reference_tuple(sheet_module: Sheet):
    from xlviews.range import reference

    ref = reference((4, 5), sheet_module)
    assert ref == f"={sheet_module.name}!$E$4"


def test_reference_error(sheet_module: Sheet):
    from xlviews.range import reference

    with pytest.raises(ValueError, match="sheet is required when `cell` is a tuple"):
        reference((4, 5))


def test_range_value_int(sheet: Sheet):
    sheet.range(1, 1).value = 10
    x = sheet.range(1, 1).value
    assert not isinstance(x, int)
    assert isinstance(x, float)
    assert x == 10


def test_range_value_str(sheet: Sheet):
    sheet.range(1, 1).value = "abc"
    x = sheet.range(1, 1).value
    assert isinstance(x, str)
    assert x == "abc"


def test_multirange_int_int(sheet_module: Sheet):
    from xlviews.range import multirange

    assert multirange(sheet_module, 3, 5).get_address() == "$E$3"


def test_multirange_error(sheet_module: Sheet):
    from xlviews.range import multirange

    with pytest.raises(TypeError):
        multirange(sheet_module, [3, 3], [5, 5])


@pytest.mark.parametrize(
    ("index", "n", "rng"),
    [
        ([3], 1, "$E$3"),
        ([(3, 5)], 3, "$E$3:$E$5"),
        ([(3, 5), 7], 4, "$E$3:$E$5,$E$7"),
        ([(3, 5), (7, 10)], 7, "$E$3:$E$5,$E$7:$E$10"),
    ],
)
def test_multirange_row(sheet_module: Sheet, index, n, rng):
    from xlviews.range import multirange

    x = multirange(sheet_module, index, 5)
    assert len(x) == n
    assert x.get_address() == rng


@pytest.mark.parametrize(
    ("index", "n", "rng"),
    [
        ([3], 1, "$C$10"),
        ([(3, 5)], 3, "$C$10:$E$10"),
        ([(3, 5), 7], 4, "$C$10:$E$10,$G$10"),
        ([(3, 5), (7, 10)], 7, "$C$10:$E$10,$G$10:$J$10"),
    ],
)
def test_multirange_column(sheet_module: Sheet, index, n, rng):
    from xlviews.range import multirange

    x = multirange(sheet_module, 10, index)
    assert len(x) == n
    assert x.get_address() == rng


@pytest.mark.parametrize(
    ("ranges", "n"),
    [
        (["A1:B3"], 6),
        (["A2:A4", "A5:A8"], 7),
        (["A2:A4,A5:A8"], 7),
        (["A2:A4,A5:A8", "C4:C7,D10:D12"], 14),
    ],
)
def test_range_collection_from_str(sheet_module: Sheet, ranges, n):
    rc = RangeCollection(ranges)
    assert len(rc) == n
    a = rc.get_address(row_absolute=False, column_absolute=False)
    assert a == ",".join(ranges)
    assert rc.first().sheet.name == sheet_module.name


@pytest.mark.parametrize(
    ("row", "n", "address"),
    [
        ([(4, 5), (10, 14)], 7, "E4:E5,E10:E14"),
        ([(5, 5), (7, 8), (10, 11)], 5, "E5,E7:E8,E10:E11"),
    ],
)
def test_range_collection_from_index_row(sheet_module: Sheet, row, n, address):
    rc = RangeCollection.from_index(sheet_module, row, 5)
    assert len(rc) == n
    a = rc.get_address(row_absolute=False, column_absolute=False)
    assert a == address


@pytest.mark.parametrize(
    ("column", "n", "address"),
    [
        ([(2, 2)], 1, "$B$5"),
        ([(4, 5), (10, 14)], 7, "$D$5:$E$5,$J$5:$N$5"),
        ([(5, 5), (7, 8), (10, 11)], 5, "$E$5,$G$5:$H$5,$J$5:$K$5"),
    ],
)
def test_range_collection_from_index_column(sheet_module: Sheet, column, n, address):
    rc = RangeCollection.from_index(sheet_module, 5, column)
    assert len(rc) == n
    assert rc.get_address() == address
    assert rc.api.Address == address


def test_range_collection_iter(sheet_module: Sheet):
    rc = RangeCollection.from_index(sheet_module, [(2, 5), (10, 12)], 1)
    for rng, row in zip(rc, [2, 3, 4, 5, 10, 11, 12], strict=True):
        assert rng.row == row


def test_range_collection_first(sheet_module: Sheet):
    rc = RangeCollection.from_index(sheet_module, [(2, 5)], 1)
    cell = rc.first()
    assert cell.row == 2
    assert cell.column == 1


def test_range_collection_repr(sheet_module: Sheet):
    rc = RangeCollection.from_index(sheet_module, [(2, 5), (8, 10)], 5)
    assert repr(rc) == "<RangeCollection $E$2:$E$5,$E$8:$E$10>"



from __future__ import annotations

from typing import TYPE_CHECKING

from xlwings import Range

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence
    from typing import Self

    from xlwings import Sheet


def reference(cell: str | tuple[int, int] | Range, sheet: Sheet | None = None) -> str:
    """Return a reference to a cell with sheet name."""
    if isinstance(cell, str):
        return cell

    if sheet is None:
        if isinstance(cell, tuple):
            raise ValueError("sheet is required when `cell` is a tuple")

        sheet = cell.sheet

    return "=" + sheet.range(*cell).get_address(include_sheetname=True)


def iter_ranges(
    sheet: Sheet,
    row: int | Sequence[int | tuple[int, int]],
    column: int | Sequence[int | tuple[int, int]],
) -> Iterator[Range]:
    if isinstance(row, int) and isinstance(column, int):
        yield sheet.range(row, column)
        return

    if isinstance(row, int) and not isinstance(column, int):
        axis = 0
        index = column
    elif isinstance(column, int) and not isinstance(row, int):
        axis = 1
        index = row
    else:
        msg = "Either row or column must be an integer."
        raise TypeError(msg)

    def get_range(start_end: int | tuple[int, int]) -> Range:
        if isinstance(start_end, int):
            start = end = start_end
        else:
            start, end = start_end

        if axis == 0:
            return sheet.range((row, start), (row, end))

        return sheet.range((start, column), (end, column))

    yield from (get_range(i) for i in index)


def union_api(ranges: Iterable[Range]):  # noqa: ANN201
    ranges = list(ranges)

    api = ranges[0].api

    if len(ranges) == 1:
        return api

    sheet = ranges[0].sheet
    union = sheet.book.app.api.Union

    for r in ranges[1:]:
        api = union(api, r.api)

    return api


def union(ranges: Iterable[Range]) -> Range:
    ranges = list(ranges)

    if len(ranges) == 1:
        return ranges[0]

    sheet = ranges[0].sheet

    return sheet.range(union_api(ranges).Address)


def multirange(
    sheet: Sheet,
    row: int | Sequence[int | tuple[int, int]],
    column: int | Sequence[int | tuple[int, int]],
) -> Range:
    """Create a discontinuous range.

    Either row or column must be an integer.
    If the other is not an integer, it is treated as a list.
    If index is (int, int), it is a simple range.
    Otherwise, each element of index is an int or (int, int), and they are
    concatenated to create a discontinuous range.

    Args:
        sheet (Sheet): The sheet object.
        row (int, tuple, or list): The row number.
        column (int, tuple, or list): The column number.

    Returns:
        Range: The discontinuous range.
    """
    return union(iter_ranges(sheet, row, column))


class RangeCollection:
    ranges: list[Range]

    def __init__(self, ranges: Iterable) -> None:
        self.ranges = []

        for rng in ranges:
            self.ranges.append(rng if isinstance(rng, Range) else Range(rng))

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        addr = self.get_address(row_absolute=True, column_absolute=True)
        return f"<{cls} {addr}>"

    @classmethod
    def from_index(
        cls,
        sheet: Sheet,
        row: int | Sequence[int | tuple[int, int]],
        column: int | Sequence[int | tuple[int, int]],
    ) -> Self:
        return cls(iter_ranges(sheet, row, column))

    def __len__(self) -> int:
        return sum(len(rng) for rng in self.ranges)

    def __iter__(self) -> Iterator[Range]:
        for rng in self.ranges:
            yield from rng

    def first(self) -> Range:
        return next(iter(self))

    def get_address(
        self,
        *,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
    ) -> str:
        return ",".join(
            rng.get_address(
                row_absolute=row_absolute,
                column_absolute=column_absolute,
                include_sheetname=include_sheetname,
                external=external,
            )
            for rng in self.ranges
        )

    @property
    def api(self):  # noqa: ANN201
        return union_api(self.ranges)
